// Dialog window handling
// Copyright (C) 2009 Fred Allen.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include "includes.h"
#import "AppInterface.h"
#import <AppKit/AppKit.h>

bool OkDialog(const char* text)
// Bring up an ok dialog, with text
// Wait for the user to click on ok, then return
// If there is a problem, SetError, and return false
{
	NSAlert* alert = [[NSAlert alloc] init];
	[alert setMessageText:[NSString stringWithCString:text encoding:NSASCIIStringEncoding]];
	[alert addButtonWithTitle:@"Ok"];

	[alert runModal];
	return (true);
}

bool OkCancelDialog(const char* text, bool* cancel)
// Bring up an ok/cancel dialog, with text
// If there is a problem, SetError, return false
// Otherwise, return true in cancel if the user
// canceled, false if not
{
	*cancel = false;

	NSAlert* alert = [[NSAlert alloc] init];
	[alert setMessageText:[NSString stringWithCString:text encoding:NSASCIIStringEncoding]];
	[alert addButtonWithTitle:@"Ok"];
	[alert addButtonWithTitle:@"Cancel"];
	NSArray* buttons = [alert buttons];
	[[buttons objectAtIndex:0] setKeyEquivalent:@"Y"];
	[[buttons objectAtIndex:1] setKeyEquivalent:@"\e"];

	NSInteger button = [alert runModal];
	if (button == NSAlertFirstButtonReturn)
	{
		*cancel = false;
	}
	else if (button == NSAlertSecondButtonReturn)
	{
		*cancel = true;
	}
	return true;
}

// ----------------------------------------------------------------------------

bool YesNoCancelDialog(const char* text, bool* yes, bool* cancel)
// Bring up a yes, no, or cancel dialog, with text
// If there is a problem, SetError, return false
// Otherwise, if cancel is true the user canceled,
// otherwise, if yes is true, yes was hit,
// otherwise no was hit.
{
	NSAlert* alert = [[NSAlert alloc] init];
	[alert setMessageText:[NSString stringWithCString:text encoding:NSASCIIStringEncoding]];
	[alert addButtonWithTitle:@"Yes"];
	[alert addButtonWithTitle:@"No"];
	[alert addButtonWithTitle:@"Cancel"];
	NSArray* buttons = [alert buttons];
	[[buttons objectAtIndex:0] setKeyEquivalent:@"y"];
	[[buttons objectAtIndex:1] setKeyEquivalent:@"n"];
	[[buttons objectAtIndex:2] setKeyEquivalent:@"\e"];

	NSInteger button = [alert runModal];
	if (button == NSAlertFirstButtonReturn)
	{
		*yes = true;
		*cancel = false;
	}
	else if (button == NSAlertSecondButtonReturn)
	{
		*yes = false;
		*cancel = false;
	}
	else if (button == NSAlertThirdButtonReturn)
	{
		*yes = false;
		*cancel = true;
	}

	return (true);
}

@implementation SimpleTextWindow
{
}

- (IBAction)okButtonAction:(id)sender
{
	[NSApp stopModalWithCode:NSModalResponseOK];
}

- (IBAction)cancelButtonAction:(id)sender
{
	[NSApp stopModalWithCode:NSModalResponseCancel];
}

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wundeclared-selector"
- (BOOL)control:(NSControl*)control textView:(NSTextView*)textView doCommandBySelector:(SEL)commandSelector
{
	BOOL result = NO;
	if (commandSelector == @selector(noop:))
	{
		NSEvent* event = [NSApp currentEvent];
		if (([event modifierFlags] & NSEventModifierFlagControl) && ([event.characters isEqualToString:@"	"] || [event.characters isEqualToString:@"\n"]))
		{
			[textView insertText:event.characters];
			result = YES;
		}
	}
	return result;
}
#pragma clang diagnostic pop
@end

// ----------------------------------------------------------------------------

NSWindow* findWindowClass(NSArray* objects)
{
	for (int i = 0; i < [objects count]; i++)
	{
		if ([objects[i] isKindOfClass:[NSWindow class]])
		{
			return objects[i];
		}
	}
	return nil;
}

bool GetSimpleTextDialog(const char* title, char* enteredText, UINT32 stringBytes, bool* cancel)
// Open a string input dialog with title,
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// otherwise, enteredText will be set to the text that the user entered in the box
// If enteredText is set to something on entry, it will be placed in the box
{
	NSArray* topObjects;
	SimpleTextWindow* sWindow;
	EDITOR_WINDOW* activeWindow = GetActiveDocumentWindow();

	if ([[NSBundle mainBundle] loadNibNamed:@"SimpleTextWindow" owner:nil topLevelObjects:&topObjects])
	{
		if ((sWindow = (SimpleTextWindow*)findWindowClass(topObjects)))
		{
			CGFloat xPos = NSWidth([[sWindow screen] frame]) / 2 - NSWidth([sWindow frame]) / 2;
			CGFloat yPos = NSHeight([[sWindow screen] frame]) / 2 - NSHeight([sWindow frame]) / 2;
			[sWindow setFrame:NSMakeRect(xPos, yPos, NSWidth([sWindow frame]), NSHeight([sWindow frame])) display:YES];

			sWindow.title = [NSString stringWithCString:title encoding:NSASCIIStringEncoding];

			sWindow.textEdit.stringValue = [NSString stringWithCString:enteredText encoding:NSASCIIStringEncoding];
			*cancel = true;
			trapCutCopyPasteForDialog = YES;
			NSInteger response = [NSApp runModalForWindow:sWindow];
			trapCutCopyPasteForDialog = NO;
			SetTopDocumentWindow(activeWindow);
			if (response == NSModalResponseOK)
			{
				if (![sWindow.textEdit.stringValue getCString:enteredText maxLength:stringBytes encoding:NSUTF8StringEncoding])
				{
					return false;
				}
				*cancel = false;
			}
			return true;
		}
	}
	return false;
}

// ----------------------------------------------------------------------------

#define kResponseFind		1000
#define kResponseFindAll	1001
#define kResponseReplace	1002
#define kResponseReplaceAll	1003

@implementation FindReplaceWindow
{
}

- (IBAction)findButtonAction:(id)sender
{
	[NSApp stopModalWithCode:kResponseFind];
}

- (IBAction)findAllButtonAction:(id)sender
{
	[NSApp stopModalWithCode:kResponseFindAll];
}

- (IBAction)replaceButtonAction:(id)sender
{
	[NSApp stopModalWithCode:kResponseReplace];
}

- (IBAction)replaceAllButtonAction:(id)sender
{
	[NSApp stopModalWithCode:kResponseReplaceAll];
}

- (IBAction)cancelButtonAction:(id)sender
{
	[NSApp stopModalWithCode:NSModalResponseCancel];
}

- (BOOL)control:(NSControl*)control textView:(NSTextView*)textView doCommandBySelector:(SEL)commandSelector
{
	BOOL result = NO;
	if (commandSelector == @selector(noop:))
	{
		NSEvent* event = [NSApp currentEvent];
		if (([event modifierFlags] & NSEventModifierFlagControl) && ([event.characters isEqualToString:@"	"] || [event.characters isEqualToString:@"\n"]))
		{
			[textView insertText:event.characters];
			result = YES;
		}
	}
	return result;
}
@end

bool SearchReplaceDialog(EDITOR_BUFFER* searchBuffer, EDITOR_BUFFER* replaceBuffer, bool* backwards, bool* wrapAround, bool* selectionExpr, bool* ignoreCase, bool* limitScope, bool* replaceProc, UINT16* searchType, bool* cancel)
// Open the search/replace dialog
// If there is a problem, return false
// Otherwise, if cancel is true, the user canceled
// Otherwise, return the enumerated mode that the user chose, as well as
// the states of backwards, wrapAround, selectionExpr, ignoreCase,
// limitScope, replaceProc
{
	NSArray* topObjects;
	FindReplaceWindow* sWindow;
	CHUNK_HEADER* theChunk;
	UINT32 bytesToExtract;
	UINT32 theOffset;
	char tmpText[8192];
	EDITOR_WINDOW* activeWindow = GetActiveDocumentWindow();

	if ([[NSBundle mainBundle] loadNibNamed:@"FindReplaceWindow" owner:nil topLevelObjects:&topObjects])
	{
		if ((sWindow = (FindReplaceWindow*)findWindowClass(topObjects)))
		{
			CGFloat xPos = NSWidth([[sWindow screen] frame]) / 2 - NSWidth([sWindow frame]) / 2;
			CGFloat yPos = NSHeight([[sWindow screen] frame]) / 2 - NSHeight([sWindow frame]) / 2;
			[sWindow setFrame:NSMakeRect(xPos, yPos, NSWidth([sWindow frame]), NSHeight([sWindow frame])) display:YES];

			[sWindow.searchBackwardCheckBox setState:(*backwards) ? NSOnState : NSOffState];
			[sWindow.wrapAroundCheckBox setState:(*wrapAround) ? NSOnState : NSOffState];
			[sWindow.limitScopeCheckBox setState:(*limitScope) ? NSOnState : NSOffState];
			[sWindow.treatReplaceTextCheckBox setState:(*replaceProc) ? NSOnState : NSOffState];
			[sWindow.selectionExpressionCheckBox setState:(*selectionExpr) ? NSOnState : NSOffState];
			[sWindow.ignoreCaseCheckBox setState:(*ignoreCase) ? NSOnState : NSOffState];


			//
			//	Set the Find text
			//
			theChunk = searchBuffer->textUniverse->firstChunkHeader;
			theOffset = 0;
			bytesToExtract = Min((sizeof(tmpText) - 1), searchBuffer->textUniverse->totalBytes);
			if (ExtractUniverseText(searchBuffer->textUniverse, theChunk, theOffset, (UINT8*)tmpText, bytesToExtract, &theChunk, &theOffset))	// read data from universe into buffer
			{
				tmpText[bytesToExtract] = '\0';
				sWindow.findTextEdit.stringValue = [NSString stringWithCString:tmpText encoding:NSASCIIStringEncoding];
			}
			//
			//	Set the replace text
			//
			theChunk = replaceBuffer->textUniverse->firstChunkHeader;
			theOffset = 0;
			bytesToExtract = Min((sizeof(tmpText) - 1), replaceBuffer->textUniverse->totalBytes);
			if (ExtractUniverseText(replaceBuffer->textUniverse, theChunk, theOffset, (UINT8*)tmpText, bytesToExtract, &theChunk, &theOffset))	// read data from universe into buffer
			{
				tmpText[bytesToExtract] = '\0';
				sWindow.replaceTextEdit.stringValue = [NSString stringWithCString:tmpText encoding:NSASCIIStringEncoding];
			}

			*cancel = true;
			trapCutCopyPasteForDialog = YES;
			NSInteger response = [NSApp runModalForWindow:sWindow];
			trapCutCopyPasteForDialog = NO;
			SetTopDocumentWindow(activeWindow);
			if (response != NSModalResponseCancel)
			{
				(*backwards) = sWindow.searchBackwardCheckBox.state == NSOnState;
				(*wrapAround) = sWindow.wrapAroundCheckBox.state == NSOnState;
				(*limitScope) = sWindow.limitScopeCheckBox.state == NSOnState;
				(*replaceProc) = sWindow.treatReplaceTextCheckBox.state == NSOnState;
				(*selectionExpr) = sWindow.selectionExpressionCheckBox.state == NSOnState;
				(*ignoreCase) = sWindow.ignoreCaseCheckBox.state == NSOnState;

				[sWindow.findTextEdit.stringValue getCString:tmpText maxLength:sizeof(tmpText) encoding:NSUTF8StringEncoding];
				EditorStartReplace(searchBuffer);
				ReplaceEditorText(searchBuffer, 0, searchBuffer->textUniverse->totalBytes, (UINT8*)tmpText, (int)strlen(tmpText), false);
				EditorEndReplace(searchBuffer);

				[sWindow.replaceTextEdit.stringValue getCString:tmpText maxLength:sizeof(tmpText) encoding:NSUTF8StringEncoding];
				EditorStartReplace(replaceBuffer);
				ReplaceEditorText(replaceBuffer, 0, replaceBuffer->textUniverse->totalBytes, (UINT8*)tmpText, (int)strlen(tmpText), false);
				EditorEndReplace(replaceBuffer);


				switch (response)
				{
					case kResponseFind:
					{
						(*searchType) = ST_FIND;
						break;
					}
					case kResponseFindAll:
					{
						(*searchType) = ST_FINDALL;
						break;
					}
					case kResponseReplace:
					{
						(*searchType) = ST_REPLACE;
						break;
					}
					case kResponseReplaceAll:
					{
						(*searchType) = ST_REPLACEALL;
						break;
					}
				}
				*cancel = false;
			}
			return true;
		}
	}
	return false;
}

// ----------------------------------------------------------------------------

@interface SimpleListBoxWindow() <NSTableViewDelegate>

@end

@implementation SimpleListBoxWindow
{
}
- (IBAction)okButtonAction:(id)sender
{
	[NSApp stopModalWithCode:NSModalResponseOK];
}

- (IBAction)cancelButtonAction:(id)sender
{
	[NSApp stopModalWithCode:NSModalResponseCancel];
}

- (NSInteger)numberOfRowsInTableView:(NSTableView*)tableView
{
	return self.numElements;
}

- (NSView*)tableView:(NSTableView*)tableView viewForTableColumn:(NSTableColumn*)tableColumn row:(NSInteger)row
{
	// Get an existing cell with the MyView identifier if it exists
	NSTextField* result = [tableView makeViewWithIdentifier:@"SimpleListBoxView" owner:self];

	// There is no existing cell to reuse so create a new one
	if (result == nil)
	{
		// Create the new NSTextField with a frame of the {0,0} with the width of the table.
		// Note that the height of the frame is not really relevant, because the row height will modify the height.
		result = [[NSTextField alloc] initWithFrame:tableView.bounds];

		// The identifier of the NSTextField instance is set to SimpleListBoxView.
		// This allows the cell to be reused.
		result.identifier = @"SimpleListBoxView";
		result.editable = NO;
		result.bordered = NO;
	}

	// result is now guaranteed to be valid, either as a reused cell
	// or as a new cell, so set the stringValue of the cell to the
	// nameArray value at row
	result.stringValue = [NSString stringWithCString:self.listElements[row] encoding:NSASCIIStringEncoding];

	// Return the result
	return result;

}

@end

bool SimpleListBoxDialog(const char* title, UINT32 numElements, const char** listElements, bool* selectedElements, bool* cancel)
// Open a list selection dialog box with two buttons (ok, cancel)
// The list dialog will be used to modify the selectionListArray, to tell
// which items were selected, and which were not
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// Otherwise, selectedElements is modified to reflect which items
// have been selected
{
	NSArray* topObjects;
	SimpleListBoxWindow* sWindow;
	EDITOR_WINDOW* activeWindow = GetActiveDocumentWindow();

	if ([[NSBundle mainBundle] loadNibNamed:@"SimpleListBoxWindow" owner:nil topLevelObjects:&topObjects])
	{
		if ((sWindow = (SimpleListBoxWindow*)findWindowClass(topObjects)))
		{
			CGFloat xPos = NSWidth([[sWindow screen] frame]) / 2 - NSWidth([sWindow frame]) / 2;
			CGFloat yPos = NSHeight([[sWindow screen] frame]) / 2 - NSHeight([sWindow frame]) / 2;
			[sWindow setFrame:NSMakeRect(xPos, yPos, NSWidth([sWindow frame]), NSHeight([sWindow frame])) display:YES];

			sWindow.title = [NSString stringWithCString:title encoding:NSASCIIStringEncoding];
			sWindow.numElements = numElements;
			sWindow.listElements = listElements;
			*cancel = true;
			trapCutCopyPasteForDialog = YES;
			NSInteger response = [NSApp runModalForWindow:sWindow];
			trapCutCopyPasteForDialog = NO;
			SetTopDocumentWindow(activeWindow);
			if (response == NSModalResponseOK)
			{
				for (int i = 0; i < numElements; i++)											// first mark every item not selected
				{
					selectedElements[i] = false;
				}

				NSUInteger currentIndex = [sWindow.tableView.selectedRowIndexes firstIndex];
				while (currentIndex != NSNotFound)
				{
					selectedElements[currentIndex] = true;
					currentIndex = [sWindow.tableView.selectedRowIndexes indexGreaterThanIndex:currentIndex];
				}
				*cancel = false;
			}
			return true;
		}
	}
	return false;
}

void FreeOpenFileDialogPaths(char** paths)
// free an array of paths that was returned from OpenFileDialog
{
	FreeStringList(paths);
}

bool OpenFileDialog(const char* title, char* fullPath, UINT32 , char** *listElements, bool* cancel)
// Bring up a file open dialog
// If there is a problem, SetError, return false
// Otherwise, cancel is true if the user canceled the dialog
//
// Otherwise, listElements contains the pointer to the start of an array of
// pointers to char which are the selections
// It must be disposed of at some later time with a call to FreeOpenFileDialogPaths
// NOTE: the array is terminated with a pointer to NULL
//
// NOTE: this is NOT allowed to change the current directory of the editor
{
	NSOpenPanel* openPanel = [NSOpenPanel openPanel];

	// Configure your panel the way you want it
	[openPanel setCanChooseFiles:YES];
	[openPanel setCanChooseDirectories:NO];
	[openPanel setAllowsMultipleSelection:YES];
	[openPanel setExtensionHidden:NO];
	[openPanel setTreatsFilePackagesAsDirectories:YES];

	[openPanel setAllowedFileTypes:nil];
	[openPanel setTitle:[NSString stringWithCString:title encoding:NSASCIIStringEncoding]];
	openPanel.directoryURL = [NSURL fileURLWithPath:[NSString stringWithCString:fullPath encoding:NSASCIIStringEncoding]];

	*cancel = true;
	trapCutCopyPasteForDialog = YES;
	NSInteger response = [openPanel runModal];
	trapCutCopyPasteForDialog = NO;
	if (response == NSModalResponseOK)
	{
		long i;
		UINT32 numElements;

		numElements = 0;
		(*listElements) = NewStringList();

		for (i = 0; i < [openPanel.URLs count]; i++)
		{
			NSString* path = openPanel.URLs[i].absoluteString;
			if ([path hasPrefix:@"file://"])
			{
				path = [path substringFromIndex:7];
			}
			const char* fullPath = [path cStringUsingEncoding:NSUTF8StringEncoding];
			AddStringToList(fullPath, listElements, &numElements);							// add it to the list
		}
		*cancel = false;
	}
	return true;
}

static int GetOffsetToFilename(char* fullPath)
{
	int index;
	bool found;
	int outIndex;

	index = (int)strlen(fullPath);
	found = false;
	outIndex = -1;
	while (!found && index > 0)
	{
		if (fullPath[index] == '/')
		{
			found = true;
		}
		else
		{
			index--;
		}
	}
	if (found)
	{
		outIndex = index + 1;
	}
	return (outIndex);
}


bool SaveFileDialog(const char* title, char* fullPath, UINT32 stringBytes, bool* cancel)
// Bring up a file save dialog
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// Otherwise, return the full path name of the chosen file
// NOTE: if the fullPath (including terminator) would be larger than stringBytes,
// then it will be truncated to fit
// NOTE ALSO: fullPath may be sent to this routine with a pathname, and the save
// dialog will open to that path (as best as it can)
//
// NOTE: this is NOT allowed to change the current directory of the editor
{
	int index;
	NSSavePanel* savePanel = [NSSavePanel savePanel];

	[savePanel setExtensionHidden:NO];
	[savePanel setAllowedFileTypes:nil];
	[savePanel setTitle:[NSString stringWithCString:title encoding:NSASCIIStringEncoding]];

	if ((index = GetOffsetToFilename(fullPath)) >= 0)
	{
		[savePanel setNameFieldStringValue:[NSString stringWithCString:&(fullPath[index]) encoding:NSASCIIStringEncoding]];

		NSString* urlPath = [NSString stringWithCString:fullPath encoding:NSASCIIStringEncoding];
		urlPath = [urlPath substringToIndex:index];
		savePanel.directoryURL = [NSURL fileURLWithPath:urlPath];
	}
	*cancel = true;
	trapCutCopyPasteForDialog = YES;
	NSInteger response = [savePanel runModal];
	trapCutCopyPasteForDialog = NO;
	if (response == NSModalResponseOK)
	{
		NSString* path = savePanel.URL.absoluteString;
		if ([path hasPrefix:@"file://"])
		{
			path = [path substringFromIndex:7];
		}
		const char* finalPath = [path cStringUsingEncoding:NSUTF8StringEncoding];
		strncpy(fullPath, finalPath, stringBytes - 1);
		*cancel = false;
	}
	return true;
}

bool ChoosePathDialog(const char* title, char* fullPath, UINT32 stringBytes, bool* cancel)
// Bring up a path choose dialog
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// Otherwise, return the full path name of the chosen path
// NOTE: if the fullPath (including terminator) would be larger than stringBytes,
// then it will be truncated to fit
// NOTE ALSO: fullPath may be sent to this routine with a pathname, and the
// dialog will open to that path (as best as it can)
//
// NOTE: this is NOT allowed to change the current directory of the editor
{
	NSOpenPanel* openPanel = [NSOpenPanel openPanel];

	// Configure your panel the way you want it
	[openPanel setCanChooseDirectories:YES];
	[openPanel setCanCreateDirectories:YES];
	[openPanel setCanChooseFiles:NO];
	[openPanel setAllowsMultipleSelection:NO];
	[openPanel setExtensionHidden:NO];

	[openPanel setAllowedFileTypes:nil];
	[openPanel setTitle:[NSString stringWithCString:title encoding:NSASCIIStringEncoding]];
	openPanel.directoryURL = [NSURL fileURLWithPath:[NSString stringWithCString:fullPath encoding:NSASCIIStringEncoding]];

	*cancel = true;
	trapCutCopyPasteForDialog = YES;
	NSInteger response = [openPanel runModal];
	trapCutCopyPasteForDialog = NO;
	if (response == NSModalResponseOK)
	{
		NSString* path = openPanel.URL.absoluteString;
		if ([path hasPrefix:@"file://"])
		{
			path = [path substringFromIndex:7];
		}
		const char* finalPath = [path cStringUsingEncoding:NSUTF8StringEncoding];
		strncpy(fullPath, finalPath, stringBytes - 1);
		*cancel = false;
	}
	return true;
}


#define	kMinFontSize	4
#define	kMaxFontSize	64

@interface ChooseFontWindow() <NSComboBoxDelegate>
@end

@implementation ChooseFontWindow
{
}

- (NSInteger)numberOfItemsInComboBox:(NSComboBox*)comboBox
{
	if (comboBox == self.familyPopupBox)
	{
		NSArray* families = [[NSFontManager sharedFontManager] availableFontFamilies];
		return [families count];
	}
	if (comboBox == self.typefacePopupBox)
	{
		return [self.familyFaces count];
	}
	if (comboBox == self.sizePopupBox)
	{
		return kMaxFontSize - kMinFontSize;
	}

	return 0;
}

- (id)comboBox:(NSComboBox*)comboBox objectValueForItemAtIndex:(NSInteger)index
{
	if (comboBox == self.familyPopupBox)
	{
		NSArray* families = [[NSFontManager sharedFontManager] availableFontFamilies];

		return families[index];
	}
	if (comboBox == self.typefacePopupBox && index < [self.familyFaces count])
	{
		return self.familyFaces[index];
	}
	if (comboBox == self.sizePopupBox && index < (kMaxFontSize - kMinFontSize))
	{
		return [NSString stringWithFormat:@"%d", index + kMinFontSize];
	}
	return @"";
}

- (void)comboBoxSelectionDidChange:(NSNotification*)notification
{
	NSComboBox* comboBox = (NSComboBox*)[notification object];

	if (comboBox == self.familyPopupBox)
	{
		NSUInteger index = [comboBox indexOfSelectedItem];
		NSArray* families = [[NSFontManager sharedFontManager] availableFontFamilies];
		NSArray* familyInfo = [[NSFontManager sharedFontManager] availableMembersOfFontFamily:families[index]];

		NSMutableArray* faces = [[NSMutableArray alloc] init];
		for (index = 0; index < [familyInfo count]; index++)
		{
			NSArray* info = familyInfo[index];
			NSString* faceName = [info[1] stringByReplacingOccurrencesOfString:@"Oblique" withString:@" Italic"];
			[faces addObject:faceName];
		}
		self.familyFaces = faces;
		[self.typefacePopupBox reloadData];
	}
	if (!self.initializing && (comboBox == self.typefacePopupBox || comboBox == self.sizePopupBox))
	{
		NSString* fName = [self makeEditorFontString];
		self.theFont = LoadEditorFont([fName cStringUsingEncoding:NSUTF8StringEncoding]);
		[self.fontDisplayLabel setFont:self.theFont->fontRef];
	}
}

- (IBAction)okButtonAction:(id)sender
{
	[NSApp stopModalWithCode:NSModalResponseOK];
}
- (IBAction)cancelButtonAction:(id)sender
{
	[NSApp stopModalWithCode:NSModalResponseCancel];
}

- (NSString*)makeEditorFontString
{
	NSUInteger familyIndex = [self.familyPopupBox indexOfSelectedItem];
	NSUInteger faceIndex = [self.typefacePopupBox indexOfSelectedItem];
	NSUInteger sizeIndex = [self.sizePopupBox indexOfSelectedItem];

	NSArray* families = [[NSFontManager sharedFontManager] availableFontFamilies];
	NSString* familyName = families[familyIndex];

	NSString* weight = @"normal";
	NSString* slant = @"r";
	NSString* face = self.familyFaces[faceIndex];
	if ([face containsString:@"Bold"])
	{
		weight = @"bold";
	}
	if ([face containsString:@"Italic"])
	{
		slant = @"i";
	}

	return [NSString stringWithFormat:@"-*-%@-%@-%@-normal--%d--75-75-*-70-*-*", familyName, weight, slant, sizeIndex + kMinFontSize];
}

@end

bool ChooseFontDialog(const char* title, char* font, UINT32 stringBytes, bool* cancel)
// Open a string input dialog with title,
// If there is a problem, SetError, return false
// Otherwise, if cancel is true, the user canceled
// otherwise, enteredText will be set to the text that the user entered in the box
// If enteredText is set to something on entry, it will be placed in the box
{
	NSArray* topObjects;
	ChooseFontWindow* sWindow;
	EDITOR_WINDOW* activeWindow = GetActiveDocumentWindow();

	if ([[NSBundle mainBundle] loadNibNamed:@"ChooseFontWindow" owner:nil topLevelObjects:&topObjects])
	{
		if ((sWindow = (ChooseFontWindow*)findWindowClass(topObjects)))
		{
			sWindow.initializing = YES;
			CGFloat xPos = NSWidth([[sWindow screen] frame]) / 2 - NSWidth([sWindow frame]) / 2;
			CGFloat yPos = NSHeight([[sWindow screen] frame]) / 2 - NSHeight([sWindow frame]) / 2;
			[sWindow setFrame:NSMakeRect(xPos, yPos, NSWidth([sWindow frame]), NSHeight([sWindow frame])) display:YES];

			sWindow.title = [NSString stringWithCString:title encoding:NSASCIIStringEncoding];
			sWindow.theFont = LoadEditorFont(font);

			[sWindow.familyPopupBox reloadData];
			sWindow.familyFaces = [[NSArray alloc] init];
			NSArray* families = [[NSFontManager sharedFontManager] availableFontFamilies];
			for (int i = 0; i < [families count]; i++)
			{
				if ([sWindow.theFont->fontRef.familyName isEqualToString:families[i]])
				{
					[sWindow.familyPopupBox selectItemAtIndex:i];
					break;
				}
			}
			[sWindow.typefacePopupBox selectItemAtIndex:0];
			NSInteger sizeIndex = sWindow.theFont->size - kMinFontSize;
			[sWindow.sizePopupBox selectItemAtIndex:sizeIndex];

			[sWindow.fontDisplayLabel setFont:sWindow.theFont->fontRef];
			sWindow.initializing = NO;
			[sWindow.familyPopupBox reloadData];
			[sWindow.sizePopupBox reloadData];
			[sWindow.typefacePopupBox reloadData];

			*cancel = true;
			NSInteger response = [NSApp runModalForWindow:sWindow];
			SetTopDocumentWindow(activeWindow);
			if (response == NSModalResponseOK)
			{
				*cancel = false;

				NSString* fName = [sWindow makeEditorFontString];
				strncpy(font, [fName cStringUsingEncoding:NSUTF8StringEncoding], stringBytes);
			}
			return true;
		}
	}
	return false;
}
