// Low level stuff needed to manage views
// Copyright (C) 2017 Fred Allen.
//
// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include "includes.h"
#import <AppKit/AppKit.h>

#define	STAGING_BUFFER_SIZE		8192												// number of bytes used in the staging buffer (must be larger than STAGING_MAX_ADD)
#define	STAGING_MAX_ADD			256													// maximum number of bytes that will ever be added to the staging buffer in one call to GetCharacterEquiv
#define MAX_SCROLL_SPEED		25

static const char hexConv[] = "0123456789ABCDEF";									// used for speedy hex conversions (for display of non-ascii codes)

// when text is being built, or widths are being calculated, these globals
// are used, they allow for much more efficient creation of text because things
// are not being passed all over for each character
static UINT8 gceStagingBuffer[STAGING_BUFFER_SIZE];									// this is where get character equivalent builds its output
static UINT32 gceBufferIndex;														// offset into the staging buffer
static UINT32 gceColumnNumber;														// running column number
static UINT16* gceWidthTable;														// width table to use while calculating positions
static UINT32 gceTabSize;															// number of spaces to use per tab
static INT32 gceRunningWidth;														// running width

static EDITOR_FONT* GetFontForStyle(VIEW_STYLE* styles, UINT32 desiredStyle)
// Return the font that should be used for the given style
// if there is none specified, then use the default
{
	if (styles[desiredStyle].font)													// if there is a font loaded for this style, return it
	{
		return (styles[desiredStyle].font);
	}
	return (styles[0].font);														// otherwise, return the default font
}

static EDITOR_COLOR GetForegroundForStyle(VIEW_STYLE* styles, UINT32 desiredStyle)
// Return the foreground color that should be used for the given style
// if there is none specified, then use the default
{
	if (styles[desiredStyle].haveForegroundColor)									// if there is a color loaded for this style, return it
	{
		return (styles[desiredStyle].foregroundColor);
	}
	return (styles[0].foregroundColor);												// otherwise, return the default color
}

static EDITOR_COLOR GetBackgroundForStyle(VIEW_STYLE* styles, UINT32 desiredStyle)
// Return the background color that should be used for the given style
// if there is none specified, then use the default
{
	if (styles[desiredStyle].haveBackgroundColor)									// if there is a color loaded for this style, return it
	{
		return (styles[desiredStyle].backgroundColor);
	}
	return (styles[0].backgroundColor);												// otherwise, return the default color
}

static bool GetCharacterEquiv(UINT8 character)
// Put the correct characters into the gceStagingBuffer at gceBufferIndex
// update gceColumnNumber, gceBufferIndex, and gceRunningWidth
// if an end of line character is seen, return false, else return true
// NOTE: this is allowed to add at most STAGING_MAX_ADD characters to the staging buffer
// When called, there must be at least that many bytes available, or
// there will be trouble
{
	UINT32 i;
	UINT32 numTabSpaces;

	if (character >= ' ' && gceWidthTable[character])									// make sure the font can display this, if not, display it as hex
	{
		gceStagingBuffer[gceBufferIndex++] = character;
		gceRunningWidth += gceWidthTable[character];
		gceColumnNumber++;
	}
	else
	{
		switch (character)
		{
			case '\n':
				return (false);
				break;
			case '\t':
				if (gceTabSize)
				{
					numTabSpaces = gceTabSize - (gceColumnNumber % gceTabSize);
					for (i = 0; i < numTabSpaces; i++)
					{
						gceStagingBuffer[gceBufferIndex++] = ' ';						// space over to tab stop
					}
					gceRunningWidth += gceWidthTable[' '] * numTabSpaces;				// add in space to running width
					gceColumnNumber += numTabSpaces;									// push to correct column
				}
				else
				{
					gceRunningWidth += gceWidthTable[gceStagingBuffer[gceBufferIndex++] = '['];
					gceRunningWidth += gceWidthTable[gceStagingBuffer[gceBufferIndex++] = hexConv[character >> 4]];
					gceRunningWidth += gceWidthTable[gceStagingBuffer[gceBufferIndex++] = hexConv[character & 0x0F]];
					gceRunningWidth += gceWidthTable[gceStagingBuffer[gceBufferIndex++] = ']'];
					gceColumnNumber += 4;
				}
				break;
			default:
				// this is a little scary looking, but it is fast
				gceRunningWidth += gceWidthTable[gceStagingBuffer[gceBufferIndex++] = '['];
				gceRunningWidth += gceWidthTable[gceStagingBuffer[gceBufferIndex++] = hexConv[character >> 4]];
				gceRunningWidth += gceWidthTable[gceStagingBuffer[gceBufferIndex++] = hexConv[character & 0x0F]];
				gceRunningWidth += gceWidthTable[gceStagingBuffer[gceBufferIndex++] = ']'];
				gceColumnNumber += 4;
				break;
		}
	}
	return (true);
}

static bool PositionAtPixel(EDITOR_VIEW* view, UINT32 linePosition, INT32 xPosition, UINT32* charPosition, UINT32* charColumn, INT32* charStartPixel, UINT32* charWidth)
// Given linePosition which points to the character at the start of a line, and xPosition
// as a "pixel" position along the line, determine the offset to the character
// at xPosition, the pixel where that character begins, and the width of that character.
// if linePosition is past EOF, charPosition will be returned as linePosition,
// charColumn will be 0, charStartPixel will be VIEW_LEFT_MARGIN, and charWidth will be 0
// if the xPosition is past the end of the line, or end of the data, this will return false
{
	GUI_VIEW_INFO* viewInfo;
	CHUNK_HEADER* chunk;
	UINT32 offset;
	bool haveStyle;
	bool inLine;
	bool loopDone;
	INT32 lastRunningWidth;
	UINT32 lastColumnNumber;
	UINT32 stylePosition;
	UINT32 numStyleElements;
	UINT32 currentStyle;

	viewInfo = view->viewInfo;															// point at the information record for this view

	PositionToChunkPosition(view->parentBuffer->textUniverse, linePosition, &chunk, &offset);

	if ((haveStyle = GetStyleRange(view->parentBuffer->styleUniverse, linePosition, &stylePosition, &numStyleElements, &currentStyle)))
	{
		ForceStyleIntoRange(currentStyle);
		stylePosition += numStyleElements;												// style will change at this position
	}

	loopDone = false;
	inLine = true;

	gceTabSize = view->tabSize;
	gceRunningWidth = lastRunningWidth = VIEW_LEFT_MARGIN;
	gceColumnNumber = lastColumnNumber = 0;
	gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);

	while (!loopDone && inLine && chunk)												// run out to the position where the text begins
	{
		if (haveStyle && (linePosition >= stylePosition))								// time to update style?
		{
			if ((haveStyle = GetStyleRange(view->parentBuffer->styleUniverse, linePosition, &stylePosition, &numStyleElements, &currentStyle)))
			{
				ForceStyleIntoRange(currentStyle);
				stylePosition += numStyleElements;										// style will change at this position
			}
			gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);
		}
		gceBufferIndex = 0;
		if ((inLine = GetCharacterEquiv(chunk->data[offset])))
		{
			if (gceRunningWidth > xPosition)
			{
				loopDone = true;
			}
			else
			{
				lastColumnNumber = gceColumnNumber;										// remember these
				lastRunningWidth = gceRunningWidth;
				linePosition++;
				offset++;
				if (offset >= chunk->totalBytes)										// see if time for a new chunk
				{
					chunk = chunk->nextHeader;
					offset = 0;
				}
			}
		}
	}
	*charPosition = linePosition;														// return the results
	*charColumn = lastColumnNumber;
	*charStartPixel = lastRunningWidth;
	*charWidth = gceRunningWidth - lastRunningWidth;

	return (chunk && inLine);
}

static void PixelAtPosition(EDITOR_VIEW* view, UINT32 position, INT32* xPosition, bool limitMax)
// Given position within view, determine the "pixel" offset to the start of the character
// of the line position is in.
// if position is passed in past the end of the buffer, this will return 0
// NOTE: if limitMax is true, this will abort as soon as the returned position
// grows larger than that passed in
{
	GUI_VIEW_INFO* viewInfo;
	CHUNK_HEADER* chunk;
	UINT32 offset;
	UINT32 linePosition;
	UINT32 line;
	UINT32 lineOffset;
	bool haveStyle;
	UINT32 stylePosition;
	UINT32 numStyleElements;
	UINT32 currentStyle;

	viewInfo = view->viewInfo;															// point at the information record for this view

	PositionToLinePosition(view->parentBuffer->textUniverse, position, &line, &lineOffset, &chunk, &offset);

	linePosition = position - lineOffset;

	if ((haveStyle = GetStyleRange(view->parentBuffer->styleUniverse, linePosition, &stylePosition, &numStyleElements, &currentStyle)))
	{
		ForceStyleIntoRange(currentStyle);
		stylePosition += numStyleElements;												// style will change at this position
	}

	gceTabSize = view->tabSize;
	gceRunningWidth = VIEW_LEFT_MARGIN;
	gceColumnNumber = 0;
	gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);

	while ((linePosition < position) && chunk && (!limitMax || (gceRunningWidth < *xPosition)))	// run out to the given position
	{
		if (haveStyle && (linePosition >= stylePosition))								// time to update style?
		{
			if ((haveStyle = GetStyleRange(view->parentBuffer->styleUniverse, linePosition, &stylePosition, &numStyleElements, &currentStyle)))
			{
				ForceStyleIntoRange(currentStyle);
				stylePosition += numStyleElements;										// style will change at this position
			}
			gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);
		}
		gceBufferIndex = 0;
		GetCharacterEquiv(chunk->data[offset]);
		linePosition++;
		offset++;
		if (offset >= chunk->totalBytes)												// see if time for a new chunk
		{
			chunk = chunk->nextHeader;
			offset = 0;
		}
	}
	(*xPosition) = gceRunningWidth;
}

void GetEditorViewGraphicToTextPosition(EDITOR_VIEW* view, UINT32 linePosition, INT32 xPosition, UINT32* betweenOffset, UINT32* charOffset)
// Given linePosition which points to the character at the start of a line, and xPosition
// as a "pixel" position along the line, determine two offsets:
// betweenOffset is the offset of the character boundary closest to the xPosition,
// charOffset is the offset of the character closest to the xPosition
// if linePosition is past EOF, both offsets will be returned as 0
{
	UINT32 charPosition;
	UINT32 charColumn;
	INT32 charStartPixel;
	UINT32 charWidth;
	bool inLine;

	inLine = PositionAtPixel(view, linePosition, xPosition, &charPosition, &charColumn, &charStartPixel, &charWidth);
	*betweenOffset = *charOffset = charPosition - linePosition;
	if (inLine && (xPosition > (charStartPixel + ((INT32)charWidth / 2))))
	{
		(*betweenOffset)++;
	}
}

void GetEditorViewTextToGraphicPosition(EDITOR_VIEW* view, UINT32 position, INT32* xPosition, bool limitMax, UINT32* slopLeft, UINT32* slopRight)
// Given position within view, determine the "pixel" offset to the start of the character
// of the line position is in.
// if position is passed in past the end of the buffer, this will return the width of the last line
// NOTE: if limitMax is true, xPosition arrives at this function with a maximum
// pixel position. This routine is then allowed to stop as soon as it
// sees the return result-slopLeft as greater than the passed value.
// NOTE: slopLeft and slopRight are used to allow fonts which have overhang
// into the previous or next character to be dealt with correctly.
// slopLeft tells how far the character at position draws to the left of
// its given position.
// slopRight tells how far the character 1 to the left of position draws
// to the right (past its character width).
{
	GUI_VIEW_INFO* viewInfo;

	viewInfo = view->viewInfo;
	*slopLeft = viewInfo->maxLeftOverhang;
	*slopRight = viewInfo->maxRightOverhang;
	if (limitMax)					// if limiting to pixel position, move out by max overhang
	{
		*xPosition += *slopLeft;
	}
	PixelAtPosition(view, position, xPosition, limitMax);
}

void GetEditorViewTextInfo(EDITOR_VIEW* view, UINT32* topLine, UINT32* numLines, INT32* leftPixel, UINT32* numPixels)
// Get information about the given view
{
	GUI_VIEW_INFO* viewInfo;

	viewInfo = view->viewInfo;
	*topLine = viewInfo->topLine;
	*numLines = viewInfo->numLines;
	*leftPixel = viewInfo->leftPixel;
	*numPixels = viewInfo->textBounds.w;
}


static bool GetViewCursorRect(EDITOR_VIEW* theView, NSRect* cursorRect)
// Create the cursor rect and return true, else return false if the cursor is not in the view
{
	GUI_VIEW_INFO* theViewInfo;
	CHUNK_HEADER* theChunk;
	UINT32 theOffset;
	UINT32 theLine;
	UINT32 theLineOffset;
	INT32 xPosition;
	INT32 maxXPosition;
	UINT32 slopLeft;
	UINT32 slopRight;

	theViewInfo = theView->viewInfo;																// point at the information record for this view

	PositionToLinePosition(theView->parentBuffer->textUniverse, GetSelectionCursorPosition(theView->parentBuffer->selectionUniverse), &theLine, &theLineOffset, &theChunk, &theOffset);
	if ((theLine >= theViewInfo->topLine) && (theLine < theViewInfo->topLine + theViewInfo->numLines))	// make sure cursor is on the view
	{
		maxXPosition = xPosition = theViewInfo->leftPixel + (theViewInfo->textBounds.w);
		GetEditorViewTextToGraphicPosition(theView, GetSelectionCursorPosition(theView->parentBuffer->selectionUniverse), &xPosition, true, &slopLeft, &slopRight);
		if ((xPosition >= theViewInfo->leftPixel) && (xPosition < maxXPosition))						// see if on view horizontally
		{
			cursorRect->origin.x = theViewInfo->textBounds.x + (xPosition - theViewInfo->leftPixel);
			cursorRect->origin.y = theViewInfo->textBounds.y + theViewInfo->lineHeight * (theLine - theViewInfo->topLine);
			cursorRect->size.width = 1;
			cursorRect->size.height = theViewInfo->lineHeight;
			return (true);
		}
	}
	return (false);
}

static void InvalidateCursorRect(EDITOR_VIEW* theView)
// Invalidate the cursor rect
{
	NSRect cursorRect;
	GUI_VIEW_INFO* theViewInfo;
	theViewInfo = theView->viewInfo;
	if (GetViewCursorRect(theView, &cursorRect))
	{
		WINDOWLISTELEMENT* theWindowElement;

		theWindowElement = (WINDOWLISTELEMENT*)theView->parentWindow->userData;
		cursorRect.origin.x = cursorRect.origin.x - theViewInfo->maxRightOverhang;										// make sure we go back far enough to include the overhang of the previous character
		cursorRect.size.width = cursorRect.size.width + theViewInfo->maxLeftOverhang + theViewInfo->maxRightOverhang;	// and go far enough forward to include the overhang of the current character and next
		[theWindowElement->viewRef setNeedsDisplayInRect:cursorRect];
	}
}

static bool CursorShowing(EDITOR_VIEW* theView)
// returns true if the cursor is showing
{
	GUI_VIEW_INFO* theViewInfo;

	theViewInfo = theView->viewInfo;
	if ((IsSelectionEmpty(theView->parentBuffer->selectionUniverse)) && theViewInfo->cursorOn && theViewInfo->blinkState && theView->active)
	{
		return (true);
	}
	return (false);
}

static void DrawViewCursor(EDITOR_VIEW* theView, CGContextRef cgContext)
// look at the cursor state for this view, and draw it as needed
// it will be drawn into the current port, and current clip shape
{
	GUI_VIEW_INFO* theViewInfo;

	theViewInfo = theView->viewInfo;			// point at the information record for this view
	if ((IsSelectionEmpty(theView->parentBuffer->selectionUniverse)) && theViewInfo->cursorOn && theView->active)
	{
		NSRect cursorRect;

		if (GetViewCursorRect(theView, &cursorRect))
		{
			if (theViewInfo->blinkState)		// Fill with either the foreground color or nothing to blink the cursor
			{
				CGContextSetRGBFillColor(cgContext, EditorColorR_float(theViewInfo->foregroundColor), EditorColorG_float(theViewInfo->foregroundColor), EditorColorB_float(theViewInfo->foregroundColor), 1);
				CGContextFillRect(cgContext, cursorRect);
			}
		}
	}
}

static void ViewStartCursorChange(EDITOR_VIEW* theView)
// when the state (not position) of the cursor
// is about to change, call this
{
	GUI_VIEW_INFO* theViewInfo;

	theViewInfo = theView->viewInfo;								// point at the information record for this view
	theViewInfo->oldCursorShowing = CursorShowing(theView);		// remember if it was showing
}

static void ViewEndCursorChange(EDITOR_VIEW* theView)
// when the state (not position) of the cursor
// has changed, call this, and it will update the cursor as needed
{
	GUI_VIEW_INFO* theViewInfo;

	theViewInfo = theView->viewInfo;								// point at the information record for this view
	if (theViewInfo->oldCursorShowing != CursorShowing(theView))	// see if the state has changed
	{
		InvalidateCursorRect(theView);
	}
}

static EDITOR_COLOR SumColors(EDITOR_COLOR colorA, EDITOR_COLOR colorB)
// When marks overlap, this is called to blend the colors of the overlapping
// marks together
{
	UINT32 red;
	UINT32 green;
	UINT32 blue;

	red = EditorColorR(colorA) + EditorColorR(colorB);
	green = EditorColorG(colorA) + EditorColorG(colorB);
	blue = EditorColorB(colorA) + EditorColorB(colorB);
	if (red > 255)
	{
		red = 255;
	}
	if (green > 255)
	{
		green = 255;
	}
	if (blue > 255)
	{
		blue = 255;
	}
	return (MakeEditorColor(red, green, blue));
}

static bool DetermineMarksBackgroundColorAtPosition(EDITOR_VIEW* view, UINT32 position, EDITOR_COLOR* color)
// Work out what the mark background color contribution is for the given position
// This looks at all visible marks, and sums their individual background color
// components. If there are no visible marks at the position, false is returned
{
	bool sawMark;
	MARK_VISIBILITY* visibleMark;

	sawMark = false;
	visibleMark = view->headVisibility;					// get head of visible marks
	while (visibleMark)
	{
		if (GetSelectionAtPosition(visibleMark->parentMark->selectionUniverse, position))	// mark active?
		{
			if (!sawMark)								// if first mark, just copy the color
			{
				*color = visibleMark->backgroundColor;
				sawMark = true;
			}
			else										// sum in other colors
			{
				*color = SumColors(*color, visibleMark->backgroundColor);
			}
		}
		visibleMark = visibleMark->nextMark;				// move to next mark hanging off of this view
	}
	return (sawMark);
}

static void DrawViewBackgroundPiece(EDITOR_VIEW* view, CGContextRef cgContext, bool selected, bool haveMarkColor, EDITOR_COLOR markColor, UINT32 theStyle, INT32 xWindowOffset, UINT32 width, INT32 yWindowOffset)
// draw the piece of text in the staging buffer into the view in the given style and selection
{
	GUI_VIEW_INFO* theViewInfo;
	EDITOR_COLOR background;

	theViewInfo = view->viewInfo;			// point at the information record for this view
	if (selected)
	{
		if (!view->haveSelectionForegroundColor && !view->haveSelectionBackgroundColor)		// if neither specified, invert
		{
			background = GetForegroundForStyle(view->viewStyles, theStyle);
		}
		else	// otherwise, do not invert
		{
			if (view->haveSelectionBackgroundColor)
			{
				background = view->selectionBackgroundColor;
			}
			else
			{
				background = GetBackgroundForStyle(view->viewStyles, theStyle);
			}
		}
	}
	else
	{
		if (!haveMarkColor)
		{
			background = GetBackgroundForStyle(view->viewStyles, theStyle);
		}
		else
		{
			background = markColor;
		}
	}
	CGContextSetRGBFillColor(cgContext, EditorColorR_float(background), EditorColorG_float(background), EditorColorB_float(background), 1);
	CGContextFillRect(cgContext, CGRectMake(xWindowOffset, yWindowOffset, width, theViewInfo->lineHeight));
}

static void DrawViewTextPiece(EDITOR_VIEW* view, CGContextRef cgContext, bool selected, bool haveMarkColor, EDITOR_COLOR markColor, UINT32 theStyle, INT32 xWindowOffset, UINT32 width, INT32 yWindowOffset)
// draw the piece of text in the staging buffer into the view in the given style and selection
{
	GUI_VIEW_INFO* theViewInfo;
	EDITOR_COLOR foreground;
	UINT32 i;

	theViewInfo = view->viewInfo;			// point at the information record for this view
	CGContextSelectFont(cgContext, GetFontForStyle(view->viewStyles, theStyle)->postScriptName, GetFontForStyle(view->viewStyles, theStyle)->size, kCGEncodingMacRoman);
	CGContextSetTextDrawingMode(cgContext, kCGTextFill);
	if (selected)
	{
		if (!view->haveSelectionForegroundColor && !view->haveSelectionBackgroundColor)		// if neither specified, invert
		{
			foreground = GetBackgroundForStyle(view->viewStyles, theStyle);
		}
		else	// otherwise, do not invert
		{
			if (view->haveSelectionForegroundColor)
			{
				foreground = view->selectionForegroundColor;
			}
			else
			{
				foreground = GetForegroundForStyle(view->viewStyles, theStyle);
			}
		}
	}
	else
	{
		foreground = GetForegroundForStyle(view->viewStyles, theStyle);
	}
	CGContextSetRGBFillColor(cgContext, EditorColorR_float(foreground), EditorColorG_float(foreground), EditorColorB_float(foreground), 1);

	// Draw a character at a time. CGContextShowTextAtPoint puts characters on fractional boundries when drawing a string, we want them on whole boundries
	// So draw a character at a time where we want them
///@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//CGContextShowTextAtPoint(cgContext, xWindowOffset, yWindowOffset + theViewInfo->lineAscent, (char*)gceStagingBuffer, gceBufferIndex);
///@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	for (i = 0; i < gceBufferIndex; i++)
	{
		CGContextShowTextAtPoint(cgContext, xWindowOffset, yWindowOffset + theViewInfo->lineAscent, (char*)&gceStagingBuffer[i], 1);
		xWindowOffset += gceWidthTable[gceStagingBuffer[i]];
	}
}

static void DrawViewLine(EDITOR_VIEW* view, CGContextRef cgContext, UINT32 currentPosition, INT32 xWindowOffset, INT32 xWindowEndOffset, INT32 yWindowOffset)
// Read bytes from the line of text at theChunk, theOffset
// Draw them onto the view
// This is responsible for drawing ALL of the pixels in the view for the line
// on which it was called (even if that line contains no text)
{
	CHUNK_HEADER* theChunk;
	UINT32 theOffset;
	GUI_VIEW_INFO* theViewInfo;
	INT32 skippedWidth;
	UINT32 currentStyle;
	UINT32 nextStyle;
	bool currentSelection;
	bool nextSelection;
	bool currentMark;
	bool nextMark;
	EDITOR_COLOR currentMarkColor;
	EDITOR_COLOR nextMarkColor;
	INT32 scrollX;																// pixel offset in text where we need to start drawing
	INT32 textXOffset;															// window relative pixel offset where text is being drawn
	INT32 distanceToStart;
	UINT32 charWidth;
	UINT32 beginningPosition;
	CHUNK_HEADER* beginningChunk;
	UINT32 beginningOffset;
	UINT32 beginningColumn;

	currentMarkColor = 0;
	nextMarkColor = 0;
	theViewInfo = view->viewInfo;													// point at the information record for this view
	scrollX = theViewInfo->leftPixel + (xWindowOffset - theViewInfo->textBounds.x);	// get pixel offset in text where we need to start drawing
	PositionAtPixel(view, currentPosition, scrollX, &currentPosition, &gceColumnNumber, &skippedWidth, &charWidth);
	PositionToChunkPosition(view->parentBuffer->textUniverse, currentPosition, &theChunk, &theOffset);

	beginningPosition = currentPosition;											// remember these for later
	beginningChunk = theChunk;
	beginningOffset = theOffset;
	beginningColumn = gceColumnNumber;

	gceBufferIndex = 0;															// begin at start of buffer
	gceRunningWidth = 0;

	distanceToStart = skippedWidth - scrollX;

	if (distanceToStart > 0)														// blank space before text starts?
	{
		if (distanceToStart > (xWindowEndOffset - xWindowOffset))					// do not go past the end of the draw area
		{
			distanceToStart = xWindowEndOffset - xWindowOffset;
		}
		CGContextSetRGBFillColor(cgContext, EditorColorR_float(view->viewStyles[0].backgroundColor), EditorColorG_float(view->viewStyles[0].backgroundColor), EditorColorB_float(view->viewStyles[0].backgroundColor), 1);
		CGContextFillRect(cgContext, CGRectMake(xWindowOffset, yWindowOffset, distanceToStart, theViewInfo->lineHeight));
	}

	textXOffset = xWindowOffset + distanceToStart;									// move the offset slightly back to beginning of first character which enters the view (or forward if view is scrolled negatively)

	currentSelection = GetSelectionAtPosition(view->parentBuffer->selectionUniverse, currentPosition);
	currentMark = DetermineMarksBackgroundColorAtPosition(view, currentPosition, &currentMarkColor);
	currentStyle = GetStyleAtPosition(view->parentBuffer->styleUniverse, currentPosition);
	ForceStyleIntoRange(currentStyle);

	gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);

	while (theChunk && ((textXOffset + gceRunningWidth) < xWindowEndOffset) && GetCharacterEquiv(theChunk->data[theOffset]))		// fill in all of the text needed
	{
		nextSelection = GetSelectionAtPosition(view->parentBuffer->selectionUniverse, currentPosition + 1);
		nextMark = DetermineMarksBackgroundColorAtPosition(view, currentPosition + 1, &nextMarkColor);
		nextStyle = GetStyleAtPosition(view->parentBuffer->styleUniverse, currentPosition + 1);
		ForceStyleIntoRange(nextStyle);

		if ((currentStyle != nextStyle) || (currentSelection != nextSelection) || (currentMark != nextMark) || (currentMark && (currentMarkColor != nextMarkColor)) || (gceBufferIndex > STAGING_BUFFER_SIZE - STAGING_MAX_ADD))	// is something about to change? (if so, output background so far) (this will also output if the staging buffer is getting full)
		{
			DrawViewBackgroundPiece(view, cgContext, currentSelection, currentMark, currentMarkColor, currentStyle, textXOffset, gceRunningWidth, yWindowOffset);

			textXOffset += gceRunningWidth;

			gceBufferIndex = 0;													// begin at start of buffer
			gceRunningWidth = 0;

			currentSelection = nextSelection;
			currentMark = nextMark;
			currentMarkColor = nextMarkColor;
			currentStyle = nextStyle;
			gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);
		}

		currentPosition++;
		theOffset++;
		if (theOffset >= theChunk->totalBytes)										// push through end of chunk
		{
			theChunk = theChunk->nextHeader;
			theOffset = 0;
		}
	}
	// ran off the end of the line, or the end of the universe
	if (gceBufferIndex)
	{
		DrawViewBackgroundPiece(view, cgContext, currentSelection, currentMark, currentMarkColor, currentStyle, textXOffset, gceRunningWidth, yWindowOffset);
		textXOffset += gceRunningWidth;
	}

	if (textXOffset < xWindowEndOffset)											// do we need to draw out to the end of the line?
	{
		if (textXOffset < xWindowOffset)											// did this end before the draw area began?
		{
			textXOffset = xWindowOffset;
		}
		currentSelection = GetSelectionAtPosition(view->parentBuffer->selectionUniverse, currentPosition);
		currentMark = DetermineMarksBackgroundColorAtPosition(view, currentPosition, &currentMarkColor);
		currentStyle = GetStyleAtPosition(view->parentBuffer->styleUniverse, currentPosition);		// get style and selection information for newline (or off of end)
		ForceStyleIntoRange(currentStyle);

		DrawViewBackgroundPiece(view, cgContext, currentSelection, currentMark, currentMarkColor, currentStyle, textXOffset, xWindowEndOffset - textXOffset, yWindowOffset);
	}


	gceBufferIndex = 0;															// begin at start of buffer
	gceRunningWidth = 0;

	currentPosition = beginningPosition;											// restore these from above
	theChunk = beginningChunk;
	theOffset = beginningOffset;
	gceColumnNumber = beginningColumn;

	textXOffset = xWindowOffset + distanceToStart;									// move the offset slightly back to beginning of first character which enters the view (or forward if view is scrolled negatively)

	currentSelection = GetSelectionAtPosition(view->parentBuffer->selectionUniverse, currentPosition);
	currentMark = DetermineMarksBackgroundColorAtPosition(view, currentPosition, &currentMarkColor);
	currentStyle = GetStyleAtPosition(view->parentBuffer->styleUniverse, currentPosition);
	ForceStyleIntoRange(currentStyle);

	gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);

	while (theChunk && ((textXOffset + gceRunningWidth) < xWindowEndOffset) && GetCharacterEquiv(theChunk->data[theOffset]))		// fill in all of the text needed
	{
		nextSelection = GetSelectionAtPosition(view->parentBuffer->selectionUniverse, currentPosition + 1);
		nextMark = DetermineMarksBackgroundColorAtPosition(view, currentPosition + 1, &nextMarkColor);
		nextStyle = GetStyleAtPosition(view->parentBuffer->styleUniverse, currentPosition + 1);
		ForceStyleIntoRange(nextStyle);

		if ((currentStyle != nextStyle) || (currentSelection != nextSelection) || (currentMark != nextMark) || (currentMark && (currentMarkColor != nextMarkColor)) || (gceBufferIndex > STAGING_BUFFER_SIZE - STAGING_MAX_ADD))	// is something about to change? (if so, output background so far) (this will also output if the staging buffer is getting full)
		{
			DrawViewTextPiece(view, cgContext, currentSelection, currentMark, currentMarkColor, currentStyle, textXOffset, gceRunningWidth, yWindowOffset);

			textXOffset += gceRunningWidth;

			gceBufferIndex = 0;													// begin at start of buffer
			gceRunningWidth = 0;

			currentSelection = nextSelection;
			currentMark = nextMark;
			currentMarkColor = nextMarkColor;
			currentStyle = nextStyle;
			gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);
		}

		currentPosition++;
		theOffset++;
		if (theOffset >= theChunk->totalBytes)										// push through end of chunk
		{
			theChunk = theChunk->nextHeader;
			theOffset = 0;
		}
	}
	// ran off the end of the line, or the end of the universe
	if (gceBufferIndex)
	{
		DrawViewTextPiece(view, cgContext, currentSelection, currentMark, currentMarkColor, currentStyle, textXOffset, gceRunningWidth, yWindowOffset);
	}
}

static void DrawViewLines(EDITOR_VIEW* theView, CGContextRef cgContext, NSRect updateRect)
// draw the lines into view, obey the invalid rect of the view
{
	GUI_VIEW_INFO* theViewInfo;
	UINT32 topLineToDraw;
	UINT32 bottomLineToDraw;
	CHUNK_HEADER* theChunk;
	UINT32 theOffset;
	UINT32 currentPosition;
	UINT32 i;
	NSRect lineRect;
	WINDOWLISTELEMENT* theWindowElement;

	theWindowElement = (WINDOWLISTELEMENT*)theView->parentWindow->userData;

	theViewInfo = theView->viewInfo;															// point at the information record for this view
	gceTabSize = theView->tabSize;
	topLineToDraw = (updateRect.origin.y - theViewInfo->textBounds.y) / theViewInfo->lineHeight;	// screen relative top line to draw
	bottomLineToDraw = (updateRect.origin.y - theViewInfo->textBounds.y + (UINT32)(updateRect.size.height) - 1) / theViewInfo->lineHeight;	// get last line to draw

	for (i = topLineToDraw; i <= bottomLineToDraw; i++)
	{
		lineRect.origin.x = updateRect.origin.x;
		lineRect.size.width = updateRect.size.width;
		lineRect.origin.y = theViewInfo->textBounds.y + (INT32)(i * theViewInfo->lineHeight);
		lineRect.size.height = theViewInfo->lineHeight;

		if ([theWindowElement->viewRef needsToDrawRect:lineRect])
		{
			LineToChunkPosition(theView->parentBuffer->textUniverse, theViewInfo->topLine + i, &theChunk, &theOffset, &currentPosition);	// point to data for this line
			DrawViewLine(theView, cgContext, currentPosition, lineRect.origin.x - theViewInfo->maxLeftOverhang, lineRect.origin.x + lineRect.size.width + theViewInfo->maxRightOverhang, lineRect.origin.y);
		}
	}
}

void DrawView(EDITOR_VIEW* theView, CGContextRef cgContext, NSRect updateRect)
// draw the given window in its parent window
// also handle highlighting of text
{
	GUI_VIEW_INFO* theViewInfo;
	CGRect wRect;
	WINDOWLISTELEMENT* theWindowElement;

	theWindowElement = (WINDOWLISTELEMENT*)theView->parentWindow->userData;
	theViewInfo = theView->viewInfo;												// point at the information record for this view

	if ((INT32)updateRect.origin.x < theViewInfo->textBounds.x)					// do we need to fill the left edge
	{
		CGContextSetRGBFillColor(cgContext, EditorColorR_float(theView->viewStyles[0].backgroundColor), EditorColorG_float(theView->viewStyles[0].backgroundColor), EditorColorB_float(theView->viewStyles[0].backgroundColor), 1);
		CGContextFillRect(cgContext, CGRectMake(theViewInfo->bounds.x, theViewInfo->bounds.y, theViewInfo->textBounds.x - theViewInfo->bounds.x, theViewInfo->bounds.h));
	}

	if ((INT32)updateRect.origin.y < theViewInfo->textBounds.y)					// do we need to fill the top
	{
		CGContextSetRGBFillColor(cgContext, EditorColorR_float(theView->viewStyles[0].backgroundColor), EditorColorG_float(theView->viewStyles[0].backgroundColor), EditorColorB_float(theView->viewStyles[0].backgroundColor), 1);
		CGContextFillRect(cgContext, CGRectMake(theViewInfo->bounds.x, theViewInfo->bounds.y, theViewInfo->bounds.w, theViewInfo->textBounds.y - theViewInfo->bounds.y));
	}

	if ((INT32)(updateRect.origin.y + updateRect.size.height) > (theViewInfo->textBounds.y + (INT32)theViewInfo->textBounds.h))	// do we need to fill the bottom
	{
		CGContextSetRGBFillColor(cgContext, EditorColorR_float(theView->viewStyles[0].backgroundColor), EditorColorG_float(theView->viewStyles[0].backgroundColor), EditorColorB_float(theView->viewStyles[0].backgroundColor), 1);
		CGContextFillRect(cgContext, CGRectMake(theViewInfo->bounds.x, theViewInfo->textBounds.y + (theViewInfo->numLines * theViewInfo->lineHeight), theViewInfo->bounds.w, theViewInfo->bounds.h - (theViewInfo->numLines * theViewInfo->lineHeight)));
	}
	wRect.origin.x = theViewInfo->textBounds.x;
	wRect.origin.y = theViewInfo->textBounds.y;
	wRect.size.width = theViewInfo->textBounds.w;
	wRect.size.height = theViewInfo->textBounds.h;

	if ([theWindowElement->viewRef needsToDrawRect:wRect])
	{
		CGRect intersectRect = CGRectIntersection(wRect, updateRect);
		DrawViewLines(theView, cgContext, intersectRect);
		DrawViewCursor(theView, cgContext);
	}
}

void RecalculateViewFontInfo(EDITOR_VIEW* view)
// recalculate the cached information for view
// from its current size and set of styles
{
	INT32 viewHeight;
	GUI_VIEW_INFO* theViewInfo;
	UINT32 i;
	EDITOR_FONT* theFont;

	theViewInfo = view->viewInfo;				// point at the information record for this view

	theViewInfo->lineHeight = 0;
	theViewInfo->lineAscent = 0;
	theViewInfo->maxLeftOverhang = 0;
	theViewInfo->maxRightOverhang = 0;
	theViewInfo->maxCharWidth = 0;

	for (i = 0; i < VIEW_MAX_STYLES; i++)
	{
		theFont = view->viewStyles[i].font;
		if (theFont != NULL)					// see if a font is assigned for this style
		{
			if (theViewInfo->lineHeight < (theFont->ascent + theFont->descent))
			{
				theViewInfo->lineHeight = theFont->ascent + theFont->descent;
				theViewInfo->lineAscent = theFont->ascent;
			}
			if (theViewInfo->maxLeftOverhang < theFont->maxLeftOverhang)
			{
				theViewInfo->maxLeftOverhang = theFont->maxLeftOverhang;
			}
			if (theViewInfo->maxRightOverhang < theFont->maxRightOverhang)
			{
				theViewInfo->maxRightOverhang = theFont->maxRightOverhang;
			}
			if (theViewInfo->maxCharWidth < theFont->maxCharWidth)
			{
				theViewInfo->maxCharWidth = theFont->maxCharWidth;
			}
		}
	}

	if (theViewInfo->lineHeight <= 0)
	{
		theViewInfo->lineHeight = 1;													// lines are minimally one pixel tall (so we do not get an infinite number of lines on the view)
	}

	theViewInfo->textBounds = theViewInfo->bounds;									// start with copy of the view bounds
	theViewInfo->textBounds.y += VIEW_TOP_MARGIN;										// leave a small margin at the top
	theViewInfo->textBounds.x += 1;													// leave 1 pixel on the left side for when we are full screen, the cursor will have a spot to be outside the text and scroll
	theViewInfo->textBounds.w -= 1;

	if (theViewInfo->bounds.h > (VIEW_TOP_MARGIN + VIEW_BOTTOM_MARGIN))
	{
		viewHeight = theViewInfo->bounds.h - VIEW_TOP_MARGIN - VIEW_BOTTOM_MARGIN;		// find out how much can be used for the text
		theViewInfo->numLines = viewHeight / theViewInfo->lineHeight;					// get integer number of lines in this view
		theViewInfo->textBounds.h = (theViewInfo->numLines * theViewInfo->lineHeight);	// alter the textBounds to be the exact rectangle to hold the text
	}
	else
	{
		theViewInfo->numLines = 0;
		theViewInfo->textBounds.w = 0;												// no room to draw text
	}
}

void SetViewBounds(EDITOR_VIEW* view, EDITOR_RECT* newBounds)
// Set the bounds of view to newBounds
// invalidate area where the view used to be but now is not
// and invalidate new areas of the view
{
	GUI_VIEW_INFO* viewInfo;
	WINDOWLISTELEMENT* theWindowElement;

	theWindowElement = (WINDOWLISTELEMENT*)view->parentWindow->userData;
	viewInfo = view->viewInfo;														// point at the information record for this view
	if ((newBounds->x != viewInfo->bounds.x) || (newBounds->y != viewInfo->bounds.y) || (newBounds->w != viewInfo->bounds.w) || (newBounds->h != viewInfo->bounds.h))
	{
		NSRect wRect;

		wRect.origin.x = viewInfo->bounds.x;
		wRect.origin.y = viewInfo->bounds.y;
		wRect.size.width = viewInfo->bounds.w;
		wRect.size.height = viewInfo->bounds.h;
		viewInfo->bounds = *newBounds;
		wRect.origin.x = viewInfo->bounds.x;
		wRect.origin.y = viewInfo->bounds.y;
		wRect.size.width = viewInfo->bounds.w;
		wRect.size.height = viewInfo->bounds.h;
		RecalculateViewFontInfo(view);												// update additional information
	}
}

void InvalidateViewPortion(EDITOR_VIEW* view, UINT32 startLine, UINT32 endLine, UINT32 startPixel, UINT32 endPixel)
// Invalidate the area of the view between startLine, endLine, and startPixel, endPixel
// NOTE: startLine, endLine, startPixel, and endPixel are all relative to the displayed view
// so that startLine of 0 always means the top line, and startPixel of 0 always
// means the left-most pixel
// NOTE: startLine must be <= endLine, startPixel must be <= endPixel
// the invalidated area is exclusive of endLine, and endPixel
{
	UINT32 totalPixels;
	GUI_VIEW_INFO* theViewInfo;
	WINDOWLISTELEMENT* theWindowElement;
	NSRect invalidRect;

	theViewInfo = view->viewInfo;													// point at the information record for this view
	theWindowElement = (WINDOWLISTELEMENT*)view->parentWindow->userData;

	totalPixels = theViewInfo->textBounds.w;
	if (theViewInfo->numLines && totalPixels)										// make sure we can actually invalidate
	{
		if (startLine > theViewInfo->numLines)
		{
			startLine = theViewInfo->numLines;
		}
		if (endLine > theViewInfo->numLines)
		{
			endLine = theViewInfo->numLines;
		}

		if (startPixel > totalPixels)
		{
			startPixel = totalPixels;
		}
		if (endPixel > totalPixels)
		{
			endPixel = totalPixels;
		}
		invalidRect.origin.x = theViewInfo->textBounds.x + startPixel;
		invalidRect.origin.y = theViewInfo->textBounds.y + startLine * theViewInfo->lineHeight;
		invalidRect.size.width = (endPixel - startPixel);
		invalidRect.size.height = ((endLine - startLine) * theViewInfo->lineHeight);
		[theWindowElement->viewRef setNeedsDisplayInRect:invalidRect];
	}
}

static void AddLineSelectionShape(EDITOR_VIEW* view, SELECTION_UNIVERSE* selection, UINT32 currentPosition, INT32 yWindowOffset)
// Read bytes from the line of text at theChunk, theOffset
// Add selected pieces to theShape
// NOTE: this must work as an exact mirror to DrawViewLine
// so that the shape accurately represents what would be selected
// if the view was drawn
{
	GUI_VIEW_INFO* viewInfo;
	CHUNK_HEADER* chunk;
	UINT32 offset;
	INT32 xWindowOffset;
	INT32 xWindowEndOffset;
	INT32 skippedWidth;
	UINT32 currentStyle;
	UINT32 nextStyle;
	bool currentSelection;
	bool nextSelection;
	INT32 scrollX;											// pixel offset in text where we need to start drawing
	INT32 textXOffset;										// window relative pixel offset where text is being drawn
	INT32 distanceToStart;
	NSRect selectionRect;
	UINT32 charWidth;
	WINDOWLISTELEMENT* theWindowElement = (WINDOWLISTELEMENT*)view->parentWindow->userData;

	viewInfo = view->viewInfo;								// point at the information record for this view

	xWindowOffset = viewInfo->textBounds.x;
	xWindowEndOffset = viewInfo->textBounds.x + viewInfo->textBounds.w;

	selectionRect.origin.y = yWindowOffset;
	selectionRect.size.height = viewInfo->lineHeight;

	scrollX = viewInfo->leftPixel;							// get pixel offset in text where we need to start drawing

	PositionAtPixel(view, currentPosition, scrollX, &currentPosition, &gceColumnNumber, &skippedWidth, &charWidth);

	PositionToChunkPosition(view->parentBuffer->textUniverse, currentPosition, &chunk, &offset);

	gceBufferIndex = 0;										// begin at start of buffer
	gceRunningWidth = 0;

	distanceToStart = skippedWidth - scrollX;

	if (distanceToStart > (xWindowEndOffset - xWindowOffset))	// do not go past the end of the draw area
	{
		distanceToStart = xWindowEndOffset - xWindowOffset;
	}

	textXOffset = xWindowOffset + distanceToStart;				// move the offset slightly back to beginning of first character which enters the view (or forward if view is scrolled negatively)

	currentSelection = GetSelectionAtPosition(selection, currentPosition);
	currentStyle = GetStyleAtPosition(view->parentBuffer->styleUniverse, currentPosition);
	ForceStyleIntoRange(currentStyle);

	gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);

	while (chunk && ((textXOffset + gceRunningWidth) < xWindowEndOffset) && GetCharacterEquiv(chunk->data[offset]))		// run through the text
	{
		nextSelection = GetSelectionAtPosition(selection, currentPosition + 1);

		nextStyle = GetStyleAtPosition(view->parentBuffer->styleUniverse, currentPosition + 1);
		ForceStyleIntoRange(nextStyle);

		if ((currentStyle != nextStyle) || (currentSelection != nextSelection) || (gceBufferIndex > STAGING_BUFFER_SIZE - STAGING_MAX_ADD))	// is something about to change? (if so, output what we have so far) (this will also output if the staging buffer is getting full)
		{
			if (currentSelection)
			{
				selectionRect.origin.x = textXOffset - viewInfo->maxLeftOverhang;
				selectionRect.size.width = gceRunningWidth + viewInfo->maxLeftOverhang + viewInfo->maxRightOverhang;
				if ((INT32)selectionRect.origin.x < xWindowOffset)						// pin to bounds
				{
					selectionRect.size.width -= (xWindowOffset - (INT32)selectionRect.origin.x);
					selectionRect.origin.x = xWindowOffset;
				}
				if ((INT32)(selectionRect.origin.x + selectionRect.size.width) > xWindowEndOffset)
				{
					selectionRect.size.width = xWindowEndOffset - (INT32)selectionRect.origin.x;
				}
				[theWindowElement->viewRef setNeedsDisplayInRect:selectionRect];
			}

			textXOffset += (INT32)gceRunningWidth;

			gceBufferIndex = 0;													// begin at start of buffer
			gceRunningWidth = 0;

			currentSelection = nextSelection;
			currentStyle = nextStyle;
			gceWidthTable = &(GetFontForStyle(view->viewStyles, currentStyle)->charWidths[0]);
		}

		currentPosition++;
		offset++;
		if (offset >= chunk->totalBytes)											// push through end of chunk
		{
			chunk = chunk->nextHeader;
			offset = 0;
		}
	}
	// ran off the end of the line, or the end of the universe
	if (gceBufferIndex)
	{
		if (currentSelection)
		{
			selectionRect.origin.x = textXOffset - viewInfo->maxLeftOverhang;
			selectionRect.size.width = gceRunningWidth + viewInfo->maxLeftOverhang + viewInfo->maxRightOverhang;
			if ((INT32)selectionRect.origin.x < xWindowOffset)							// pin to bounds
			{
				selectionRect.size.width -= (xWindowOffset - (INT32)selectionRect.origin.x);
				selectionRect.origin.x = xWindowOffset;
			}
			if ((INT32)(selectionRect.origin.x + selectionRect.size.width) > xWindowEndOffset)
			{
				selectionRect.size.width = xWindowEndOffset - (INT32)selectionRect.origin.x;
			}
			[theWindowElement->viewRef setNeedsDisplayInRect:selectionRect];
		}
		textXOffset += gceRunningWidth;
	}

	if (textXOffset < xWindowEndOffset)											// do we need to go out to the end of the line?
	{
		if (textXOffset < xWindowOffset)											// did this end before the draw area began?
		{
			textXOffset = xWindowOffset;
		}
		currentSelection = GetSelectionAtPosition(selection, currentPosition);		// get selection information for newline (or off of end)

		if (currentSelection)
		{
			selectionRect.origin.x = textXOffset;
			selectionRect.size.width = (INT32)selectionRect.origin.x + (xWindowEndOffset - textXOffset);
			if ((INT32)selectionRect.origin.x < xWindowOffset)							// pin to bounds
			{
				selectionRect.size.width -= (xWindowOffset - (INT32)selectionRect.origin.x);
				selectionRect.origin.x = xWindowOffset;
			}
			if ((INT32)(selectionRect.origin.x + selectionRect.size.width) > xWindowEndOffset)
			{
				selectionRect.size.width = (xWindowEndOffset - (INT32)selectionRect.origin.x);
			}
			[theWindowElement->viewRef setNeedsDisplayInRect:selectionRect];
		}
	}
}

void InvalidateViewSelection(EDITOR_VIEW* view, SELECTION_UNIVERSE* selection)
// Passed a view, and a selection, invalidate the visible area of the view
// covered by selection
{
	GUI_VIEW_INFO* viewInfo;
	CHUNK_HEADER* chunk;
	UINT32 offset;
	UINT32 currentPosition;
	UINT32 i;
	WINDOWLISTELEMENT* theWindowElement;

	viewInfo = view->viewInfo;									// point at the information record for this view
	theWindowElement = (WINDOWLISTELEMENT*)view->parentWindow->userData;

	gceTabSize = view->tabSize;
	for (i = 0; i < viewInfo->numLines; i++)
	{
		LineToChunkPosition(view->parentBuffer->textUniverse, viewInfo->topLine + i, &chunk, &offset, &currentPosition);	// point to data for line
		AddLineSelectionShape(view, selection, currentPosition, viewInfo->textBounds.y + i * viewInfo->lineHeight);
	}
}

void InvalidateView(EDITOR_VIEW* view)
// invalidate the entire area of the view
{
	WINDOWLISTELEMENT* theWindowElement = (WINDOWLISTELEMENT*)view->parentWindow->userData;
	theWindowElement->viewRef.needsDisplay = YES;
}

void ScrollViewPortion(EDITOR_VIEW* view, UINT32 startLine, UINT32 endLine, INT32 numLines, UINT32 startPixel, UINT32 endPixel, INT32 numPixels)
// Scroll the area of the view between startLine, endLine, and startPixel, endPixel
// by numLines, and numPixels
// If numLines > 0, the data is scrolled down, else it is scrolled up
// If numPixels > 0, the data is scrolled right, else it is scrolled left
// any area left open by the scroll will be left erased, and invalidated
// NOTE: startLine, endLine, startPixel, and endPixel are all relative to the displayed view
// so that startLine of 0 always means the top line, and startPixel of 0 always
// means the left-most pixel
// NOTE: startLine must be <= endLine, startPixel must be <= endPixel
// the scrolled area is exclusive of endLine, and endPixel
{
	UINT32 totalPixels;
	NSRect sourceRect;
	NSRect clipRect;
	GUI_VIEW_INFO* theViewInfo;
	WINDOWLISTELEMENT* theWindowElement;

	theViewInfo = view->viewInfo;											// point at the information record for this view
	theWindowElement = (WINDOWLISTELEMENT*)view->parentWindow->userData;

	totalPixels = theViewInfo->textBounds.w;
	if (theViewInfo->numLines && totalPixels)								// make sure we can actually scroll
	{
		if (startLine > theViewInfo->numLines)
		{
			startLine = theViewInfo->numLines;
		}
		if (endLine > theViewInfo->numLines)
		{
			endLine = theViewInfo->numLines;
		}

		if (startPixel > totalPixels)
		{
			startPixel = totalPixels;
		}
		if (endPixel > totalPixels)
		{
			endPixel = totalPixels;
		}

		if (numLines > ((INT32)(endLine - startLine)))						// allow a scroll that erases everything
		{
			numLines = endLine - startLine;
		}
		else
		{
			if (numLines < ((INT32)startLine - (INT32)endLine))
			{
				numLines = (INT32)startLine - (INT32)endLine;
			}
		}

		if (numPixels > ((INT32)(endPixel - startPixel)))
		{
			numPixels = endPixel - startPixel;
		}
		else
		{
			if (numPixels < (INT32)startPixel - (INT32)endPixel)
			{
				numPixels = (INT32)startPixel - (INT32)endPixel;
			}
		}
		if (numLines || numPixels)
		{
			sourceRect.origin.x = theViewInfo->textBounds.x + startPixel;
			sourceRect.origin.y = theViewInfo->textBounds.y + startLine * theViewInfo->lineHeight;
			sourceRect.size.width = endPixel - startPixel;
			sourceRect.size.height = (endLine - startLine) * theViewInfo->lineHeight;

			clipRect.origin.x = theViewInfo->textBounds.x;
			clipRect.origin.y = theViewInfo->textBounds.y;
			clipRect.size.width = theViewInfo->textBounds.w;
			clipRect.size.height = theViewInfo->textBounds.h;

			sourceRect = CGRectIntersection(clipRect, sourceRect);
			if (!CGRectIsEmpty(sourceRect))
			{
				NSSize delta = CGSizeMake(numPixels, numLines * (INT32)theViewInfo->lineHeight);
				[theWindowElement->viewRef scrollRect:sourceRect by:delta];
			}
		}
	}
}

void SetViewTopLeft(EDITOR_VIEW* theView, UINT32 newTopLine, INT32 newLeftPixel)
// Set the upper left corner of view
// NOTE: newTop, and newLeft will be forced into bounds by this routine
{
	UINT32 oldTopLine;
	INT32 oldLeftPixel;
	UINT32 numLines;
	GUI_VIEW_INFO* theViewInfo;
	WINDOWLISTELEMENT* theWindowElement;

	theWindowElement = (WINDOWLISTELEMENT*)theView->parentWindow->userData;
	theViewInfo = theView->viewInfo;											// point at the information record for this view
	numLines = theView->parentBuffer->textUniverse->totalLines;
	oldTopLine = theViewInfo->topLine;
	oldLeftPixel = theViewInfo->leftPixel;

	if (newTopLine > numLines)
	{
		newTopLine = numLines;
	}
	if (theViewInfo->topLine != newTopLine || theViewInfo->leftPixel != newLeftPixel)
	{
		UpdateEditorWindows();												// make sure the display is up to date before changing it
		theViewInfo->topLine = newTopLine;
		theViewInfo->leftPixel = newLeftPixel;
		ScrollViewPortion(theView, 0, theViewInfo->numLines, oldTopLine - newTopLine, 0, theViewInfo->textBounds.w, oldLeftPixel - newLeftPixel);
		if (theView->viewPositionChangedVector)								// see if someone needs to be called because of my change
		{
			theView->viewPositionChangedVector(theView);					// call him with my pointer
		}
		UpdateEditorWindows();												// redraw now to make the update more crisp
	}
}

void SetEditorViewTopLine(EDITOR_VIEW* view, UINT32 lineNumber)
// Set the line number of the top line that is
// displayed in the given edit view
// NOTE: this does no invalidation of the view
{
	view->viewInfo->topLine = lineNumber;
}

void BlinkViewCursor(EDITOR_VIEW* theView)
// calling this will cause the cursor in theView to change blink states, updating it as needed
{
	GUI_VIEW_INFO* theViewInfo;

	theViewInfo = theView->viewInfo;											// point at the information record for this view
	ViewStartChange(theView);
	theViewInfo->blinkState = !theViewInfo->blinkState;						// invert blink state

	InvalidateCursorRect(theView);
	ViewEndChange(theView);
}

void ResetEditorViewCursorBlink(EDITOR_VIEW* view)
// When this is called, it will reset the cursor blinking, force the
// cursor to the blink-on state, and reset the blink timeout
{
	if (!view->viewInfo->blinkState)											// if it is currently off, blink it to on
	{
		BlinkViewCursor(view);
	}
	ResetCursorBlinkTime(view);
}

void ViewCursorTask(EDITOR_VIEW* view)
// Handle blinking the cursor for view
{
	BlinkViewCursor(view);
}

void ViewStyleFontsChanged(EDITOR_VIEW* view)
// When the fonts have been modified in styles for the passed view,
// this is called so that it may update the line height, and anything else
// that may be necessary.
{
	RecalculateViewFontInfo(view);
}

void ViewEndChange(EDITOR_VIEW* view)
// The passed view has finished being changed. Bring the view back
// up to date
{
	ViewStartCursorChange(view);
	view->viewInfo->cursorOn = true;											// turn cursor back on when done
	ViewEndCursorChange(view);
}

void ViewStartChange(EDITOR_VIEW* view)
// The passed view is about to be changed in some way, do whatever is needed
// so that when ViewEndChange is called, the correct parts of the view are
// invalidated.
{
	ViewStartCursorChange(view);
	view->viewInfo->cursorOn = false;											// turn cursor back on when done
	ViewEndCursorChange(view);
}

void DisposeViewInfo(EDITOR_VIEW* view)
// Dispose of the view
{
	MDisposePtr(view->viewInfo);
}

bool CreateViewInfo(EDITOR_VIEW* view, EDITOR_VIEW_DESCRIPTOR* descriptor)
// Called after all editor specific fields of a new view have been filled in
// This creates and stuffs local GUI specific information, and sets up the view in the GUI
// if there is a problem, SetError, and return false
{
	if ((view->viewInfo = (GUI_VIEW_INFO*)MNewPtr(sizeof(GUI_VIEW_INFO))))
	{
		view->viewInfo->topLine = descriptor->topLine;
		view->viewInfo->leftPixel = descriptor->leftPixel;		// initial pixel to display
		view->viewInfo->bounds = descriptor->rect;				// copy bounds rectangle

		view->viewInfo->cursorOn = true;							// the cursor is turned on
		view->viewInfo->blinkState = false;						// next time it blinks, it will blink to on
		RecalculateViewFontInfo(view);							// update view information based on fonts and stuff
		return (true);
	}
	return (false);
}


void OSXClickPosToViewClickPos(EDITOR_VIEW* theView, INT32 xPos, INT32 yPos, VIEW_CLICK_EVENT_DATA* theClickData)
// Convert a OS X click position relative to the current client area, to a click position relative to the current view.
// This routine clips the click position so it's always inside the view.
{
	GUI_VIEW_INFO* theViewInfo;

	theViewInfo = (GUI_VIEW_INFO*)theView->viewInfo;
	xPos -= theViewInfo->textBounds.x;
	yPos -= theViewInfo->textBounds.y;
	if (yPos < 0)
	{
		yPos -= theViewInfo->lineHeight;
	}
	yPos = yPos / ((INT32)(theViewInfo->lineHeight));
	theClickData->xClick = xPos;
	theClickData->yClick = yPos;
}

@implementation GUIView
{
}

- (id)initWithFrame:(NSRect)frame
{
	self = [super initWithFrame:frame];
	if (self)
	{
		[self registerForDraggedTypes:[NSArray arrayWithObjects:NSColorPboardType, NSFilenamesPboardType, nil]];
	}
	return self;
}

- (NSDragOperation)draggingEntered:(id <NSDraggingInfo>)sender
{
	NSPasteboard* pboard;
	NSDragOperation sourceDragMask;

	sourceDragMask = [sender draggingSourceOperationMask];
	pboard = [sender draggingPasteboard];

	if ([[pboard types] containsObject:NSFilenamesPboardType])
	{
		if (sourceDragMask & NSDragOperationLink)
		{
			return NSDragOperationLink;
		}
	}
	return NSDragOperationNone;
}

- (BOOL)performDragOperation:(id <NSDraggingInfo>)sender
{
	NSPasteboard* pboard;
	NSDragOperation sourceDragMask;

	sourceDragMask = [sender draggingSourceOperationMask];
	pboard = [sender draggingPasteboard];

	if ([[pboard types] containsObject:NSFilenamesPboardType])
	{
		if (sourceDragMask & NSDragOperationLink)
		{
			NSArray* files = [pboard propertyListForType:NSFilenamesPboardType];

			char** theList;
			UINT32 numElements = 0;
			theList = NewStringList();
			for (int i = 0; i < [files count]; i++)
			{
				const char* fullPath = [files[i] UTF8String];
				AddStringToList(fullPath, &theList, &numElements);											// add that to our list
			}
			if (numElements)
			{
				HandleShellCommand("open", (int)numElements, (const char**)theList);															// tell Tcl script to open these files
			}
			FreeStringList(theList);
		}
	}
	return YES;
}

- (BOOL)acceptsFirstResponder
{
	return YES;
}

- (BOOL)isFlipped
{
	return true;
}

- (BOOL)isOpaque
{
	return YES;
}


- (void)handleMouseDownEvent:(NSEvent*)theEvent
// We just got a mouse down, handle it. This will send the mouse down to the editor and then continue to send VET_CLICKHOLD(basically mouse moves) until we
// get a mouse up.
{
	UINT32 scrollSpeed;
	UINT32 scrollSpeedX;
	UINT32 scrollSpeedY;

	WINDOWLISTELEMENT* theWindowElement;
	GUI_VIEW_INFO* theViewInfo;
	EDITOR_VIEW* theView;
	EDITOR_WINDOW* theWindow;
	VIEW_CLICK_EVENT_DATA mouseClickData;
	VIEW_EVENT mouseViewEvent;
	UInt32 numClicks;

	theWindowElement = self.windowElement;
	theWindow = theWindowElement->theEditorWindow;
	theView = (EDITOR_VIEW*)theWindow->windowInfo;
	theViewInfo = theView->viewInfo;

	numClicks = (UInt32)[theEvent clickCount];
	numClicks = (numClicks - 1) % 4;																			// numclicks in 1-n range, convert to 0-3 range for editor

	mouseClickData.modifiers = OSXToEditorModifiers([theEvent modifierFlags]);									// convert OS X modifiers to editors
	mouseClickData.keyCode = (UINT32)[theEvent buttonNumber] * 2;												// left or right mouse button
	switch ([theEvent buttonNumber])
	{
		case 0:
		{
			mouseClickData.keyCode = 0;																				// left mouse button
			if ([theEvent modifierFlags] & NSEventModifierFlagControl)												// on single button mouses, if the control key is press
			{
				mouseClickData.keyCode = 2;																			// select right mouse button. Should we query they system and only do this for single mouse buttons?
			}
			break;
		}
		case 1:
		{
			mouseClickData.keyCode = 2;																				// right mouse button
			break;
		}
		case 2:
		{
			mouseClickData.keyCode = 1;																				// middle mouse button
			break;
		}
	}

	mouseViewEvent.eventType = VET_CLICK;
	mouseViewEvent.view = theView;
	mouseViewEvent.eventData = &mouseClickData;
	numClicks <<= EES_STATE0;
	numClicks &= EEM_STATE0;
	mouseClickData.modifiers |= numClicks;
	NSPoint event_location = [theEvent locationInWindow];
	NSPoint local_point = [self convertPoint:event_location fromView:nil];
	OSXClickPosToViewClickPos(theView, (INT32)local_point.x, (INT32)local_point.y, &mouseClickData);			// convert from window to text view coords.
	HandleViewEvent(&mouseViewEvent);																			// send the mouse down

	scrollSpeed = 1;

	BOOL keepOn = YES;
	NSPoint mouseLoc = [self convertPoint:[theEvent locationInWindow] fromView:nil];

	while (keepOn)
	{
		theEvent = [NSApp nextEventMatchingMask:NSLeftMouseUpMask | NSLeftMouseDraggedMask | NSRightMouseUpMask | NSRightMouseDraggedMask | NSOtherMouseUpMask | NSOtherMouseDraggedMask
					untilDate:[[NSDate alloc] initWithTimeIntervalSinceNow:0.0025]
					inMode:NSDefaultRunLoopMode
					dequeue:YES];

		switch ([theEvent type])
		{
			case NSOtherMouseDragged:
			case NSRightMouseDragged:
			case NSLeftMouseDragged:
			{
				mouseLoc = [self convertPoint:[theEvent locationInWindow] fromView:nil];							// get a new mouse location and fall through
			}
			default:
			{
				scrollSpeedY = 0;																					// assume fast scroll to start with
				if (mouseLoc.y < 0)																					// if the cursor is above the window (scrolling up)
				{
					scrollSpeedY = abs((int)mouseLoc.y);															// then the value is just the Abs(y)
				}
				else if (mouseLoc.y > theViewInfo->textBounds.h)													// is the cursor below the window
				{
					scrollSpeedY = mouseLoc.y - theViewInfo->textBounds.h;											// yes, scroll down. See how much below the window
				}
				else
				{
				    [[NSCursor IBeamCursor] set];
				}
				scrollSpeedY = Min(scrollSpeedY, MAX_SCROLL_SPEED);																// don't go past MAX_SCROLL_SPEED
				scrollSpeedY = MAX_SCROLL_SPEED - scrollSpeedY;																	// the value right now is 0-MAX_SCROLL_SPEED, with MAX_SCROLL_SPEED being further away, scroll faster. We need to it be 0 scroll fast, MAX_SCROLL_SPEED slow

				//
				// for the X, for now just go fast if the cursor is left or right of the window and not doing Y,
				// else slow and let the Y determine the true speed
				//
				if (scrollSpeedY < MAX_SCROLL_SPEED && (mouseLoc.x < 0 || mouseLoc.x > theViewInfo->textBounds.w))
				{
					scrollSpeedX = 1;
				}
				else
				{
					scrollSpeedX = MAX_SCROLL_SPEED;
				}
				scrollSpeed = Min(scrollSpeedX, scrollSpeedY);

				OSXClickPosToViewClickPos(theView, (INT32)mouseLoc.x, (INT32)mouseLoc.y, &mouseClickData);
				mouseViewEvent.eventType = VET_CLICKHOLD;
				mouseViewEvent.view = theView;
				mouseViewEvent.eventData = &mouseClickData;
				HandleViewEvent(&mouseViewEvent);
				break;
			}
			case NSRightMouseUp:
			case NSLeftMouseUp:
			{
				keepOn = NO;																						// mouse up, break out of the drag loop
				break;
			}
		}
	}
}

- (void)scrollWheel:(NSEvent*)theEvent
{
	WINDOWLISTELEMENT* theWindowElement;
	GUI_VIEW_INFO* theViewInfo;
	EDITOR_VIEW* theView;
	EDITOR_WINDOW* theWindow;

	theWindowElement = self.windowElement;
	theWindow = theWindowElement->theEditorWindow;
	theView = (EDITOR_VIEW*)theWindow->windowInfo;
	theViewInfo = theView->viewInfo;

	int i;
	int deltaY = theEvent.scrollingDeltaY;
	if (deltaY > 0)																									// scroll up?
	{
		for (i = 0; i < deltaY; i++)																						// call scroll up this many times. would be nice to have a scroll this many lines call
		{
			ScrollEditViewEvent(theView, VVK_SCROLLUP, 0);
		}
	}
	else
	{
		deltaY = abs(deltaY);																						// scrolling down, make the value positive
		for (i = 0; i < deltaY; i++)
		{
			ScrollEditViewEvent(theView, VVK_SCROLLDOWN, 0);
		}
	}
}

- (void)otherMouseDown:(NSEvent*)theEvent
{
//@@@@@@@@ work on this, middle mouse button
//	[self handleMouseDownEvent:theEvent];
}

- (void)rightMouseDown:(NSEvent*)theEvent
{
	[self handleMouseDownEvent:theEvent];
}

- (void)mouseDown:(NSEvent*)theEvent
{
	[self handleMouseDownEvent:theEvent];
}

- (void)mouseEntered:(NSEvent *)theEvent
{
	[super mouseEntered:theEvent];
    [[NSCursor IBeamCursor] set];
}

- (void)mouseExited:(NSEvent *)theEvent
{
	[super mouseExited:theEvent];
	[[NSCursor arrowCursor] set];
}

- (void)mouseMoved:(NSEvent *)theEvent;
{
	[super mouseMoved:theEvent];
    [[NSCursor IBeamCursor] set];
}

- (void)resetCursorRects
{
	[super resetCursorRects];
	[self removeTrackingArea:self.trackingArea];
	self.trackingArea = [[NSTrackingArea alloc] initWithRect:[self visibleRect] options:(NSTrackingMouseEnteredAndExited | NSTrackingMouseMoved | NSTrackingActiveInKeyWindow) owner:self userInfo:nil];
	[self addTrackingArea:self.trackingArea];

	NSCursor* iBeamCursor = [NSCursor IBeamCursor];
	[self discardCursorRects];
	[self addCursorRect:self.bounds cursor:iBeamCursor];
	[iBeamCursor setOnMouseEntered:YES];
}

- (BOOL)handleKeyDownEvent:(NSEvent*)event keyEquivalent:(BOOL)keyEquiv
{
//NSLog(@"%@", event);
	WINDOWLISTELEMENT* theWindowElement;
	GUI_VIEW_INFO* theViewInfo;
	EDITOR_VIEW* theView;
	EDITOR_WINDOW* theWindow;
	EDITOR_KEY theKeyData;
	VIEW_EVENT theViewEvent;
	INT32 osxKey;

	theWindowElement = self.windowElement;
	theWindow = theWindowElement->theEditorWindow;
	theView = (EDITOR_VIEW*)theWindow->windowInfo;
	theViewInfo = theView->viewInfo;

	unichar keyChar = 0;
	NSString* keyCharStr = [event charactersIgnoringModifiers];
	if ([keyCharStr length] != 0)
	{
		keyChar = [keyCharStr characterAtIndex:0];
	}

	theKeyData.modifiers = OSXToEditorModifiers([event modifierFlags]);										// convert OS X modifiers to editors
	theKeyData.isVirtual = true;
	switch (event.keyCode)
	{
		case kVK_F1:
		{
			theKeyData.keyCode = VVK_F1;
			break;
		}
		case kVK_F2:
		{
			theKeyData.keyCode = VVK_F2;
			break;
		}
		case kVK_F3:
		{
			theKeyData.keyCode = VVK_F3;
			break;
		}
		case kVK_F4:
		{
			theKeyData.keyCode = VVK_F4;
			break;
		}
		case kVK_F5:
		{
			theKeyData.keyCode = VVK_F5;
			break;
		}
		case kVK_F6:
		{
			theKeyData.keyCode = VVK_F6;
			break;
		}
		case kVK_F7:
		{
			theKeyData.keyCode = VVK_F7;
			break;
		}
		case kVK_F8:
		{
			theKeyData.keyCode = VVK_F8;
			break;
		}
		case kVK_F9:
		{
			theKeyData.keyCode = VVK_F9;
			break;
		}
		case kVK_F10:
		{
			theKeyData.keyCode = VVK_F10;
			break;
		}
		case kVK_F11:
		{
			theKeyData.keyCode = VVK_F11;
			break;
		}
		case kVK_F12:
		{
			theKeyData.keyCode = VVK_F12;
			break;
		}
		case kVK_F13:
		{
			theKeyData.keyCode = VVK_F13;
			break;
		}
		case kVK_F14:
		{
			theKeyData.keyCode = VVK_F14;
			break;
		}
		case kVK_F15:
		{
			theKeyData.keyCode = VVK_F15;
			break;
		}
		case kVK_F16:
		{
			theKeyData.keyCode = VVK_F16;
			break;
		}
		case kVK_F17:
		{
			theKeyData.keyCode = VVK_F17;
			break;
		}
		case kVK_F18:
		{
			theKeyData.keyCode = VVK_F18;
			break;
		}
		case kVK_F19:
		{
			theKeyData.keyCode = VVK_F19;
			break;
		}
		case kVK_F20:
		{
			theKeyData.keyCode = VVK_F20;
			break;
		}
		case kVK_Home:
		{
			theKeyData.keyCode = VVK_HOME;
			break;
		}
		case kVK_ANSI_KeypadEnter:						// there's no e93 virtual key code, but we do want to take this and see if a key is bound to it
		{
			break;
		}
		case kVK_End:
		{
			theKeyData.keyCode = VVK_END;
			break;
		}
		case kVK_Delete:
		{
			theKeyData.keyCode = VVK_BACKSPACE;
			break;
		}
		case kVK_Escape:
		{
			theKeyData.keyCode = VVK_ESC;
			break;
		}
		case kVK_Tab:
		{
			theKeyData.keyCode = VVK_TAB;
			break;
		}
		case kVK_PageUp:
		{
			theKeyData.keyCode = VVK_PAGEUP;
			break;
		}
		case kVK_PageDown:
		{
			theKeyData.keyCode = VVK_PAGEDOWN;
			break;
		}
		case kVK_Return:
		{
			theKeyData.keyCode = VVK_RETURN;
			break;
		}
		case kVK_LeftArrow:
		{
			theKeyData.keyCode = VVK_LEFTARROW;
			break;
		}
		case kVK_RightArrow:
		{
			theKeyData.keyCode = VVK_RIGHTARROW;
			break;
		}
		case kVK_UpArrow:
		{
			theKeyData.keyCode = VVK_UPARROW;
			break;
		}
		case kVK_DownArrow:
		{
			theKeyData.keyCode = VVK_DOWNARROW;
			break;
		}
		case kVK_ForwardDelete:
		{
			theKeyData.keyCode = VVK_DELETE;
			break;
		}
		case kVK_Help:
		{
			theKeyData.keyCode = VVK_INSERT;
			break;
		}
		default:
		{
			osxKey = FindOSXVirtualKey(event.keyCode);
			if (osxKey != -1)
			{
				if (HandleBoundKeyEvent(osxKey, theKeyData.modifiers))
				{
					return YES;
				}
			}
			theKeyData.isVirtual = false;
			break;
		}
	}
	if (theKeyData.isVirtual)
	{
		osxKey = FindOSXVirtualKey(event.keyCode);
		if (osxKey != -1 && HandleBoundKeyEvent(osxKey, theKeyData.modifiers))
		{
			return YES;
		}
		else
		{
			theViewEvent.view = theView;						// point at the view
			theViewEvent.eventType = VET_KEYDOWN;				// Key down event
			theViewEvent.eventData = &theKeyData;
			HandleViewEvent(&theViewEvent);						// pass this off to high level handler
			return YES;
		}
	}
	else if (!keyEquiv && keyChar > 0 && keyChar < 256)			// make sure it is 1-255, ascii
	{
		theKeyData.keyCode = keyChar;
		theViewEvent.view = theView;
		theViewEvent.eventType = VET_KEYDOWN;
		theViewEvent.eventData = &theKeyData;
		HandleViewEvent(&theViewEvent);							// pass this off to high level handler
		return YES;
	}
	return NO;
}

- (BOOL)performKeyEquivalent:(NSEvent *)event
{
	return [self handleKeyDownEvent:event keyEquivalent:YES];
}

- (void)keyDown:(NSEvent*)event
{
	[self handleKeyDownEvent:event keyEquivalent:NO];
}

- (void)drawRect:(NSRect)rect
{
	WINDOWLISTELEMENT* theWindowElement;
	GUI_VIEW_INFO* theViewInfo;
	EDITOR_VIEW* theView;
	EDITOR_WINDOW* theWindow;

	theWindowElement = self.windowElement;
	theWindow = theWindowElement->theEditorWindow;
	theView = (EDITOR_VIEW*)theWindow->windowInfo;
	theViewInfo = theView->viewInfo;

	NSGraphicsContext* gc = [NSGraphicsContext currentContext];
	CGContextRef context = (CGContextRef)[gc graphicsPort];

	CGContextSetShouldAntialias(context, true);						// we want nice looking text
	CGContextSetAllowsAntialiasing(context, true);
	CGContextSetShouldSmoothFonts(context, true);
	CGContextSetTextMatrix(context, CGAffineTransformMake(1.0, 0.0, 0.0, -1.0, 0.0, 0.0));
	CGContextSetTextDrawingMode(context, kCGTextFill);

	DrawView(theView, context, rect);
}
@end

